"""Deep learning layers"""
from typing import Optional, Sequence

import tensorflow as tf
import tensorflow.keras.backend as kb
from tensorflow.keras.layers import Layer
from megnet.layers.readout import Set2Set
from megnet.utils.layer import repeat_with_index


class WeightedAverageLayer(Layer):
    r"""
    Weight average the features using weights

    result= \sum{w_i^a * value_i} / \sum{w_i^a}

    """

    def __init__(self, alpha: float = 1.0, **kwargs):
        """
        Args:
            alpha (float): exponent in weighting
        """
        super().__init__(**kwargs)
        self.alpha = alpha

    def build(self, input_shape: Sequence) -> None:
        """
        Build the layer

        Args:
            input_shape (tuple): input shape tuple
        """
        self.built = True

    def call(self, inputs: Sequence, mask: Optional[tf.Tensor] = None) -> tf.Tensor:
        """
        Core logic of the layer

        Args:
            inputs (tuple): input tuple of length 3
            mask (tf.Tensor): not used here
        """
        # prop: [1, n, n_feature]
        # weight: [1, n]
        # index： [1, n]
        prop, weights, index = inputs
        expo_weights = weights ** self.alpha
        prop = prop * expo_weights[:, :, None]
        prop_sum = self.reduce_sum(prop, index, perm=[1, 0, 2])
        weight_sum = self.reduce_sum(expo_weights, index, perm=[1, 0])
        return prop_sum / weight_sum[:, :, None]

    @staticmethod
    def reduce_sum(prop: tf.Tensor, index: tf.Tensor, perm: Sequence) -> tf.Tensor:
        """
        Reduce sum the tensors using index

        Args:
            prop (tf.Tensor): tensor with shape [1, n, ...]
            index (tf.Tensor): integer tensor with shape [1, n]
            perm (list): permutation for transpose.
        """
        index = tf.reshape(index, (-1,))
        prop = tf.transpose(a=prop, perm=perm)
        out = tf.math.segment_sum(prop, index)
        out = tf.transpose(a=out, perm=perm)
        return out

    @staticmethod
    def compute_output_shape(input_shape: Sequence) -> tuple:
        """
        Compute output shape from input shape
        Args:
            input_shape (tuple/list): input shape list
        """
        prop_shape = input_shape[0]
        return prop_shape[0], None, prop_shape[-1]

    def get_config(self) -> dict:
        """
        Get layer configuration
        """
        config = {"alpha": self.alpha}
        base_config = super().get_config()
        config.update(base_config)
        return config  # type: ignore


class WeightedSet2Set(Set2Set):
    """
    This is a modified version from megnet.layers.readout.Set2Set.
    Here, in addition to taking features and indices as inputs, we also
    take a weight tensor. The input to the core logic is
    [features, weights, indices]
    """

    def build(self, input_shape: Sequence) -> None:
        """
        Build the output shape from input shapes
        Args:
            input_shape (tuple/list): input shape list
        """
        feature_shape, weight_shape, index_shape = input_shape
        super().build([feature_shape, index_shape])

    def call(self, inputs, mask=None):
        """
        Core logic of the layer

        Args:
            inputs (tuple): input tuple of length 3
            mask (tf.Tensor): not used here
        """
        features, weights, feature_graph_index = inputs
        feature_graph_index = tf.reshape(feature_graph_index, (-1,))
        _, _, count = tf.unique_with_counts(feature_graph_index)
        m = kb.dot(features, self.m_weight)
        if self.use_bias:
            m += self.m_bias

        self.h = tf.zeros(tf.stack([tf.shape(input=features)[0], tf.shape(input=count)[0], self.n_hidden]))
        self.c = tf.zeros(tf.stack([tf.shape(input=features)[0], tf.shape(input=count)[0], self.n_hidden]))
        q_star = tf.zeros(tf.stack([tf.shape(input=features)[0], tf.shape(input=count)[0], 2 * self.n_hidden]))
        for i in range(self.T):
            self.h, c = self._lstm(q_star, self.c)
            e_i_t = tf.reduce_sum(input_tensor=m * repeat_with_index(self.h, feature_graph_index), axis=-1)
            exp = tf.exp(e_i_t) * weights
            # print('exp shape ', exp.shape)
            seg_sum = tf.transpose(
                a=tf.math.segment_sum(tf.transpose(a=exp, perm=[1, 0]), feature_graph_index), perm=[1, 0]
            )
            seg_sum = tf.expand_dims(seg_sum, axis=-1)
            # print('seg_sum shape', seg_sum.shape)
            interm = repeat_with_index(seg_sum, feature_graph_index)
            # print('interm shape', interm.shape)
            a_i_t = exp / interm[..., 0]
            # print(a_i_t.shape)
            r_t = tf.transpose(
                a=tf.math.segment_sum(
                    tf.transpose(a=tf.multiply(m, a_i_t[:, :, None]), perm=[1, 0, 2]), feature_graph_index
                ),
                perm=[1, 0, 2],
            )
            q_star = kb.concatenate([self.h, r_t], axis=-1)
        return q_star



"""Model mixins"""

from typing import Union, Optional, List

import numpy as np
from maml.utils import DummyScaler
from ._model import BaseModel


class TargetScalerMixin:
    """
    Mixin class for models with target scaler
    """

    def __init__(self, *args, **kwargs):
        """
        Mixin with target scaler, required kwarg target_scaler

        Args:
            *args: input args for model constructor
            **kwargs: kwargs for model constructor
        """
        target_scaler = kwargs.pop("target_scaler", DummyScaler())
        self.target_scaler = target_scaler
        super().__init__(*args, **kwargs)

    def train(
        self,
        objs: Union[List, np.ndarray],
        targets: Optional[Union[List, np.ndarray]] = None,
        val_objs: Optional[Union[List, np.ndarray]] = None,
        val_targets: Optional[Union[List, np.ndarray]] = None,
        **kwargs,
    ) -> "BaseModel":
        """
        Train the models from object, target pairs

        Args:
            objs (list of objects): List of objects
            targets (list): list of float or np.ndarray
            val_objs (list of objects): list of validation objects
            val_targets (list): list of validation targets
            **kwargs:

        Returns: self

        """
        if targets is None:
            return super().train(objs, targets, val_objs, val_targets, **kwargs)  # type: ignore

        num_atoms = [len(i) for i in objs]
        targets = [self.target_scaler.transform(i, j) for i, j in zip(targets, num_atoms)]
        if val_objs is not None and val_targets is not None:
            num_val_atoms = [len(i) for i in val_objs]
            val_targets = [self.target_scaler.transform(i, j) for i, j in zip(val_targets, num_val_atoms)]
        super().train(objs, targets, val_objs, val_targets, **kwargs)  # type: ignore
        return self  # type: ignore

    def predict_objs(self, objs: Union[List, np.ndarray]) -> np.ndarray:
        """
        Predict the values given a set of objects. Usually Pymatgen
            Structure objects.
        """
        preds = super().predict_objs(objs)  # type: ignore
        ns = [len(i) for i in objs]
        return np.array([self.target_scaler.inverse_transform(i, j) for i, j in zip(preds, ns)])



"""
Wrapper for matminer_wrapper featurizers
"""

import logging
from inspect import signature
from typing import Any, Callable, Optional

import pandas as pd

from maml.base import BaseDescriber

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def wrap_matminer_describer(
    cls_name: str, wrapped_class: Any, obj_conversion: Callable, describer_type: Optional[Any] = None
):
    """
    Wrapper of matminer_wrapper describers.
    Args:
        cls_name (str): new class name
        wrapped_class (class object): matminer_wrapper BaseFeaturizer
        obj_conversion (callable): function to convert objects into desired
            object type within transform_one
        describer_type (object): object type

    Returns: maml describers class
    """

    def constructor(self, *args, **kwargs):
        """
        Wrapped __init__ constructor
        """
        n_jobs = kwargs.pop("n_jobs", 0)
        memory = kwargs.pop("memory", None)
        verbose = kwargs.pop("verbose", False)
        feature_batch = kwargs.pop("feature_concat", "pandas_concat")
        wrapped_class.__init__(self, *args, **kwargs)
        logger.info(f"Using matminer_wrapper {wrapped_class.__name__} class")
        base_kwargs = dict(n_jobs=n_jobs, memory=memory, verbose=verbose, feature_batch=feature_batch)
        BaseDescriber.__init__(self, **base_kwargs)

    @classmethod  # type: ignore
    def _get_param_names(cls):  # type: ignore
        return wrapped_class._get_param_names()

    def get_params(self, deep=False):
        params = wrapped_class.get_params(self, deep=deep)
        return params

    def transform_one(self, obj: Any):
        """
        featurize to transform_one
        """
        obj = obj_conversion(obj)
        results = wrapped_class.featurize(self, obj)
        labels = wrapped_class.feature_labels(self)
        return pd.DataFrame({i: [j] for i, j in zip(labels, results)})

    @classmethod  # type: ignore
    def from_preset(cls, name: str, **kwargs):  # type: ignore
        """
        Wrap matminer_wrapper's from_preset function
        """
        instance = wrapped_class.from_preset(name)
        sig = signature(wrapped_class.__init__)
        args = list(sig.parameters.keys())[1:]
        params = {i: None for i in args}
        params.update(**kwargs)
        instance_new = cls(**params)
        instance_new.__dict__.update(instance.__dict__)
        return instance_new

    new_klass = type(
        cls_name,
        (BaseDescriber,),
        {
            "__doc__": wrapped_class.__doc__,
            "__init__": constructor,
            "__str__": wrapped_class.__str__,
            "__repr__": wrapped_class.__repr__,
            "__getstate__": wrapped_class.__getstate__,
            "__setstate__": wrapped_class.__setstate__,
            "_get_param_names": _get_param_names,
            "transform_one": transform_one,
            "from_preset": from_preset,
            "get_params": get_params,
            "__module__": "maml.describers",
            "describer_type": describer_type,
        },
    )

    return new_klass



"""
maml - materials machine learning
"""

__version__ = "0.0.4"

from .base import BaseDataSource  # noqa
from .base import BaseDescriber  # noqa
from .base import BaseModel, SKLModel, KerasModel, SequentialDescriber  # noqa



"""
Utils for describers
"""

import logging
from collections import Counter
from functools import partial
from typing import List, Optional, Tuple

import numpy as np

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _add_allowed_stats(cls):
    """
    Decorate to add allowed_stats to the Stats class

    Args:
        cls: Stats class

    Returns: Stats class with allowed_stats attributes

    """
    all_keys = list(cls.__dict__.keys())
    allowed = []
    for key in all_keys:
        if isinstance(cls.__dict__[key], staticmethod):
            allowed.append(key)
    setattr(cls, "allowed_stats", allowed)
    return cls


@_add_allowed_stats
class Stats:
    """
    Calculate the stats of a list of values.
    This is particularly helpful when you want to convert
    lists of values of different lengths to uniform length
    for machine learning purposes.

    supported
    """

    @staticmethod
    def max(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Max of value
        Args:31
            data (list): list of float data
            weights (list): optional weights

        Returns: maximum value

        """
        return np.max(data)

    @staticmethod
    def min(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        min of value
        Args:
            data (list): list of float data
            weights (list): optional weights

        Returns: minimum value

        """
        return np.min(data)

    @staticmethod
    def range(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Range of values
        Args:
            data (list): list of float data
            weights (list): optional weights

        Returns: range of values, i.e., max - min

        """
        return Stats.max(data) - Stats.min(data)

    @staticmethod
    def mode(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Mode of data, if multiple entries have equal counts,
        compute the average of those

        Args:
            data (list): list of float data
            weights (list): optional weights

        Returns: mode of values, i.e., max - min

        """

        if weights is None:
            counts = Counter(data)
            most_common = counts.most_common()
            max_count = most_common[0][1]
            modes = []
            for v, c in most_common:
                if c == max_count:
                    modes.append(v)
            return np.mean(modes).item()

        data_array = np.array(data)
        weights_array = np.array(weights)
        maxes = np.isclose(weights_array, weights_array.max())
        max_data = data_array[maxes]
        return np.mean(max_data).item()

    @staticmethod
    def mean_absolute_deviation(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        mean absolute deviation

        Args:
            data (list): list of float data
            weights (list): optional weights

        Returns: mean absolute deviation
        """
        mean = Stats.mean(data, weights)
        data_sub = [abs(i - mean) for i in data]
        return Stats.mean(data_sub, weights)

    @staticmethod
    def mean_absolute_error(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        mean absolute error

        Args:
            data (list): list of float data
            weights (list): optional weights

        Returns: mean absolute error
        """
        return Stats.mean_absolute_deviation(data=data, weights=weights)

    @staticmethod
    def moment(
        data: List[float],
        weights: Optional[List[float]] = None,
        order: Optional[int] = None,
        max_order: Optional[int] = None,
    ):
        """
        Moment of probability mass function

        order = 1 means weighted mean
        order = 2 means standard deviation
        order > 2 corresponds to higher order moment to
            the 1./order power

        Args:
            data (list): list of float data
            weights (list or None): weights for each data points
            order (int): moment order
            max_order (int): if set, it will overwrite order

        Returns: float or list of floats

        """
        # check if only a single value should be output
        single = max_order is None

        if weights is None:
            weights = [1.0 / len(data)] * len(data)

        if max_order is not None:
            if order is not None:
                logger.info("max_order will overwrite the order parameter")
            orders = list(range(1, max_order + 1))
        else:
            orders = [order or 1]

        stats = [_root_moment(data, weights, i) for i in orders]

        if not single:
            return stats
        return stats[0]

    @staticmethod
    def mean(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Weighted average

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: average value

        """
        return Stats.moment(data, weights=weights, order=1)

    @staticmethod
    def inverse_mean(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        inverse mean

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: average value

        """
        return Stats.mean([1.0 / x for x in data], weights=weights)

    @staticmethod
    def average(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Weighted average

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: average value

        """
        return Stats.mean(data, weights=weights)

    @staticmethod
    def std(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Standard deviation

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: Standard deviation

        """

        return Stats.moment(data, weights=weights, order=2)

    @staticmethod
    def skewness(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Skewness of the distribution

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: Skewness of the distribution

        """
        third = Stats.moment(data, weights=weights, order=3)
        std = Stats.std(data, weights=weights)
        if std < 1e-4:
            std = 1.0
        return third ** 3 / std ** 3

    @staticmethod
    def kurtosis(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Kurtosis of the distribution

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: Kurtosis of the distribution

        """
        fourth = Stats.moment(data, weights=weights, order=4)
        std = Stats.std(data, weights=weights)
        if std < 1e-4:
            std = 1.0
        return fourth ** 4 / std ** 4

    @staticmethod
    def geometric_mean(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        Geometric mean of the data

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: geometric mean of the distribution

        """
        return Stats.power_mean(data, weights, p=0)

    @staticmethod
    def power_mean(data: List[float], weights: Optional[List[float]] = None, p: int = 1) -> float:
        """
        power mean https://en.wikipedia.org/wiki/Generalized_mean

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point
            p (int): power

        Returns: power mean of the distribution

        """

        if np.any(np.array(data) <= 0.0):
            raise ValueError("Not possible to calculate geometric means for negative values")

        if weights is None:
            weights = [1.0 / len(data)] * len(data)

        else:
            weights = [i / sum(weights) for i in weights]

        assert abs(sum(weights) - 1) < 1e-3

        if p == 0:
            return np.prod([i ** j for i, j in zip(data, weights)]).item()

        s = np.sum([j * i ** p for i, j in zip(data, weights)])
        return s ** (1.0 / p)

    @staticmethod
    def shifted_geometric_mean(data: List[float], weights: Optional[List[float]] = None, shift: float = 100) -> float:
        """
        Since we cannot calculate the geometric means on negative or zero values,
        we can first shift all values to positive and then calculate the geometric mean
        afterwards, we shift the computed geometric mean back by a shift value

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point
            shift (float): shift value

        Returns: geometric mean of the distribution

        """
        data_new = [i + shift for i in data]
        return Stats.geometric_mean(data_new, weights=weights) - shift

    @staticmethod
    def harmonic_mean(data: List[float], weights: Optional[List[float]] = None) -> float:
        """
        harmonic mean of the data

        Args:
            data (list): list of float data
            weights (list or None): weights for each data point

        Returns: harmonic mean of the distribution

        """
        if weights is None:
            weights = [1.0 / len(data)] * len(data)
        else:
            weights = [i / sum(weights) for i in weights]
        return np.sum(weights) / np.sum(np.array(weights) / np.array(data))


def _root_moment(data, weights, order) -> float:
    """
    Auxilliary function to compute moment

    Args:
        data (list): list of float data
        weights (list or None): weights for each data point
        order (int): order of moment

    Returns: moment of order

    """
    weights_sum = np.sum(weights)
    pmf = np.array(weights) / weights_sum
    mean = 0.0

    if order > 1:
        mean = np.mean(data)

    moment = sum([(i - mean) ** order * j for i, j in zip(data, pmf)])

    # when order is odd, moment can be negative
    if moment < 0:
        sign = -1
    else:
        sign = 1

    # avoid error like np.power(-0.001, 1./3.)
    res = (sign * moment) ** (1.0 / order) * sign
    return res


def _convert_a_or_b(v: str, a=int, b=None):
    try:
        return a(v)
    except ValueError:
        return b


def _moment_symbol_conversion(moment_symbol: str):
    splits = moment_symbol.split(":")
    max_order = _convert_a_or_b(splits[2], int, None)

    if max_order is None:
        return [moment_symbol]
    if max_order > 0:
        return ["moment:%d:None" % i for i in range(1, max_order + 1)]
    return ["moment:0:None"]


def stats_list_conversion(stats_list: List[str]) -> List[str]:
    """
    Convert a list of stats str into a fully expanded list.
    This applies mainly to stats that can return a list of values, e.g.,
    moment with max_order > 1

    Args:
        stats_list (list): list of stats str

    Returns: list of expanded stats str

    """
    re_list = []
    for st in stats_list:
        if ":" not in st:
            re_list.append(st)
        elif "moment" in st:
            moment = _moment_symbol_conversion(st)
            re_list.extend(moment)
        else:
            re_list.append(st)
    return re_list


STATS_KWARGS = {"moment": [{"order": int}, {"max_order": int}], "shifted_geometric_mean": [{"shift": float}]}


def get_full_stats_and_funcs(stats: List) -> Tuple[List[str], List]:
    """
    Get expanded stats function name str and the corresponding
    function callables

    Args:
        stats (list): a list of stats names, e.g, ['mean', 'std', 'moment:1:None']

    Returns: list of stats names, list of stats callable

    """
    stats_func = []
    full_stats = stats_list_conversion(stats)

    for stat in full_stats:
        if ":" in stat:
            splits = stat.split(":")
            stat_name = splits[0]

            if stat_name.lower() not in getattr(Stats, "allowed_stats", []):  # type: ignore
                raise ValueError(f"{stat_name.lower()} not in available Stats")

            func = getattr(Stats, stat_name)
            args = splits[1:]
            arg_dict = {}
            for name_dict, arg in zip(STATS_KWARGS[stat_name], args):  # type: ignore
                name = list(name_dict.keys())[0]
                value_type = list(name_dict.values())[0]
                try:
                    value = value_type(arg)
                except ValueError:
                    value = None  # type: ignore
                arg_dict[name] = value
            stats_func.append(partial(func, **arg_dict))
            continue

        if stat.lower() not in getattr(Stats, "allowed_stats", []):  # type: ignore
            raise ValueError(f"{stat.lower()} not in available Stats")
        stats_func.append(getattr(Stats, stat))
    return full_stats, stats_func



"""
neural network models
"""
import math
from typing import Optional, Sequence, Union, List, Any

import numpy as np

from maml.base import KerasModel, BaseDescriber, BaseModel


def construct_atom_sets(
    input_dim: Optional[int] = None,
    is_embedding: bool = True,
    n_neurons: Sequence[int] = (64, 64),
    n_neurons_final: Sequence[int] = (64, 64),
    n_targets: int = 1,
    activation: str = "relu",
    embedding_vcal: int = 95,
    embedding_dim: int = 32,
    symmetry_func: Union[List[str], str] = "mean",
    optimizer: str = "adam",
    loss: str = "mse",
    compile_metrics: tuple = (),
    is_classification: bool = False,
    **symmetry_func_kwargs,
):
    r"""
    f(X) = \rho(\sum_{x \in X} \phi(x)), where X is a set.
    \phi is implemented as a neural network and \rho is a symmetry function.

    todo: implement attention mechanism

    Args:
        input_dim (int): input dimension, if None, then integer inputs + embedding are assumed.
        is_embedding (bool): whether the input should be embedded
        n_neurons (tuple): number of hidden-layer neurons before passing to symmetry function
        n_neurons_final (tuple): number of hidden-layer neurons after symmetry function
        n_targets (int): number of output targets
        activation (str): activation function
        embedding_vcal (int): int, embedding vocabulary
        embedding_dim (int): int, embedding dimension
        symmetry_func (str): symmetry function, choose from ['set2set', 'sum', 'mean',
            'max', 'min', 'prod']
        optimizer (str): optimizer for the models
        loss (str): loss function for the models
        compile_metrics (tuple): metrics for validation
        symmetry_func_kwargs (dict): kwargs for symmetry function
    """
    from tensorflow.keras.layers import Input, Dense, Embedding, Concatenate
    from tensorflow.keras.models import Model

    if is_embedding and input_dim is not None:
        raise ValueError("When embedding is used, input dim needs to be None")

    if is_embedding:
        inp = Input(shape=(None,), dtype="int32", name="node_id")
        out_ = Embedding(embedding_vcal, embedding_dim)(inp)
    else:
        inp = Input(shape=(None, input_dim), dtype="float32", name="node_feature_input")
        out_ = inp

    weight_inputs = Input(shape=(None,), dtype="float32", name="weight_input")
    node_ids = Input(shape=(None,), dtype="int32", name="node_in_graph_id")

    # start neural networks \phi
    for n_neuron in n_neurons:
        out_ = Dense(n_neuron, activation=activation)(out_)

    # apply symmetry function \rho
    if isinstance(symmetry_func, str):
        symmetry_func = [symmetry_func]

    symmetry_layers = []
    for symm in symmetry_func:
        if symm == "set2set":
            from maml.models.dl._layers import WeightedSet2Set

            layer = WeightedSet2Set(**symmetry_func_kwargs)
        elif symm == "mean":
            from maml.models.dl._layers import WeightedAverageLayer

            alpha = symmetry_func_kwargs.pop("alpha", 1)
            layer = WeightedAverageLayer(alpha=alpha)
        else:
            raise ValueError("symmetry function not supported")
        symmetry_layers.append(layer)
    outs = [i([out_, weight_inputs, node_ids]) for i in symmetry_layers]

    if len(outs) > 1:
        out_ = Concatenate(axis=-1)(outs)
    else:
        out_ = outs[0]

    # neural networks
    for n_neuron in n_neurons_final:
        out_ = Dense(n_neuron, activation=activation)(out_)

    if is_classification:
        final_act: Optional[str] = "sigmoid"
    else:
        final_act = None
    out_ = Dense(n_targets, activation=final_act)(out_)
    model = Model(inputs=[inp, weight_inputs, node_ids], outputs=out_)
    model.compile(optimizer, loss, metrics=compile_metrics)
    return model


class AtomSets(KerasModel):
    r"""
    This class implements the DeepSets models
    """

    def __init__(
        self,
        describer: Optional[BaseDescriber] = None,
        input_dim: Optional[int] = None,
        is_embedding: bool = True,
        n_neurons: Sequence[int] = (64, 64),
        n_neurons_final: Sequence[int] = (64, 64),
        n_targets: int = 1,
        activation: str = "relu",
        embedding_vcal: int = 95,
        embedding_dim: int = 32,
        symmetry_func: Union[List[str], str] = "mean",
        optimizer: str = "adam",
        loss: str = "mse",
        compile_metrics: tuple = (),
        is_classification: bool = False,
        **symmetry_func_kwargs,
    ):
        """
        Args:
            describer (BaseDescriber): site describers
            input_dim (int): input dimension, if None, then integer inputs + embedding are assumed.
            is_embedding (bool): whether the input should be embedded
            n_neurons (tuple): number of hidden-layer neurons before passing to symmetry function
            n_neurons_final (tuple): number of hidden-layer neurons after symmetry function
            n_targets (int): number of output targets
            activation (str): activation function
            embedding_vcal (int): int, embedding vocabulary
            embedding_dim (int): int, embedding dimension
            symmetry_func (str): symmetry function, choose from ['set2set', 'sum', 'mean',
                'max', 'min', 'prod']
            optimizer (str): optimizer for the models
            loss (str): loss function for the models
            symmetry_func_kwargs (dict): kwargs for symmetry function

        """
        input_dim = self.get_input_dim(describer, input_dim)
        model = construct_atom_sets(
            input_dim=input_dim,
            is_embedding=is_embedding,
            n_neurons=n_neurons,
            n_neurons_final=n_neurons_final,
            n_targets=n_targets,
            activation=activation,
            embedding_vcal=embedding_vcal,
            embedding_dim=embedding_dim,
            symmetry_func=symmetry_func,
            optimizer=optimizer,
            loss=loss,
            compile_metrics=compile_metrics,
            is_classification=is_classification,
            **symmetry_func_kwargs,
        )
        self.is_embedding = is_embedding
        super().__init__(model=model, describer=describer)

    def _get_data_generator(self, features, targets, batch_size=128, is_shuffle=True):
        if features is None:
            return None
        from tensorflow.keras.utils import Sequence as KerasSequence

        def _generate_atom_indices(lengths):
            max_length = max(lengths)
            res = np.tile(np.arange(max_length)[None, :], (len(lengths), 1))
            res2 = np.tile(np.arange(len(lengths))[:, None], (1, max_length))
            return res2[res < np.array(lengths)[:, None]]

        is_embedding = self.is_embedding

        class _DataGenerator(KerasSequence):
            def __init__(self, features=features, targets=targets, batch_size=batch_size, is_shuffle=is_shuffle):
                if isinstance(features[0], list) and len(features[0]) == 2:
                    self.features = [i[0] for i in features]
                    self.weights = [i[1] for i in features]
                else:
                    self.features = features
                    self.weights = [np.ones(shape=(len(i),)) for i in features]
                self.targets = targets
                self.batch_size = batch_size
                self.is_shuffle = is_shuffle

            def __len__(self):
                return math.ceil(len(self.features) / self.batch_size)

            def __getitem__(self, idx):
                features_temp = self.features[idx * self.batch_size : (idx + 1) * self.batch_size]
                batch_x = np.concatenate(features_temp, axis=0)
                if is_embedding:
                    batch_x = batch_x[..., 0]
                lengths = [len(i) for i in features_temp]
                f_index = _generate_atom_indices(lengths)
                batch_weights = self.weights[idx * self.batch_size : (idx + 1) * self.batch_size]
                batch_y = self.targets[idx * self.batch_size : (idx + 1) * self.batch_size]
                return (
                    np.array(batch_x)[None, ...],
                    np.concatenate(batch_weights)[None, :],
                    f_index[None, :],
                ), np.array(batch_y)[None, :]

            def on_epoch_end(self):
                """
                Codes executed at the end of each epoch
                """
                if self.is_shuffle:
                    indices = list(range(len(self.features)))
                    np.random.shuffle(indices)
                    self.features = [self.features[i] for i in indices]
                    self.weights = [self.weights[i] for i in indices]
                    self.targets = [self.targets[i] for i in indices]

        return _DataGenerator()

    def fit(
        self,
        features: Union[List, np.ndarray],
        targets: Union[List, np.ndarray] = None,
        val_features: Optional[Union[List, np.ndarray]] = None,
        val_targets: Optional[Union[List, np.ndarray]] = None,
        **kwargs,
    ) -> "BaseModel":
        """
        Args:
            features (list or np.ndarray): Numerical input feature list or
                numpy array with dim (m, n) where m is the number of data and
                n is the feature dimension.
            targets (list or np.ndarray): Numerical output target list, or
                numpy array with dim (m, ).
            val_features (list or np.ndarray): validation features
            val_targets (list or np.ndarray): validation targets

        """
        batch_size = kwargs.pop("batch_size", 128)
        is_shuffle = kwargs.pop("is_shuffle", True)
        train_generator = self._get_data_generator(features, targets, batch_size=batch_size, is_shuffle=is_shuffle)
        val_generator = self._get_data_generator(
            val_features, val_targets, batch_size=batch_size, is_shuffle=is_shuffle
        )
        return self.model.fit(train_generator, validation_data=val_generator, **kwargs)  # type: ignore

    def _predict(self, features: np.ndarray, **kwargs) -> np.ndarray:
        """
        Predict the values given a set of inputs based on fitted models.

        Args:
            features (np.ndarray): array-like input features.

        Returns:
            List of output objects.
        """
        predict_generator = self._get_data_generator(features, [0] * len(features), is_shuffle=False, **kwargs)
        predicted = []
        for batch in predict_generator:
            predicted.append(self.model.predict(batch[0])[0])  # type: ignore
        return np.concatenate(predicted, axis=0)

    def evaluate(self, eval_objs: Any, eval_targets: Any, is_feature: bool = False):
        """
        Evaluate objs, targets

        Args:
            eval_objs (list): objs for evaluation
            eval_targets (list): target list for the corresponding objects
            is_feature (bool): whether x is feature matrix
        """
        eval_features = eval_objs if is_feature else self.describer.transform(eval_objs)
        eval_generator = self._get_data_generator(eval_features, eval_targets)
        return self.model.evaluate(eval_generator)



"""
Selectors
"""
import inspect
from collections import defaultdict
from itertools import combinations
from typing import List, Optional, Union, Dict, Callable

import numpy as np
from scipy.linalg import lstsq
from scipy.optimize import minimize, NonlinearConstraint
from sklearn.linear_model import LinearRegression
from sklearn.metrics import get_scorer

from joblib import Parallel, delayed


# pylint: disable=R0201
class BaseSelector:
    """
    Feature selector. This is meant to work on relatively smaller
    number of features
    """

    def __init__(self, coef_thres: float = 1e-6, method: str = "SLSQP"):
        """
        Base selector
        Args:
            coef_thres (float): threshold to discard certain coefficents
            method (str): optimization methods in scipy.optmize.minimize
        """
        self.coef_thres = coef_thres
        self.is_fitted = False
        self.coef_: Optional[np.ndarray] = None
        self.method = method
        self.indices: Optional[np.ndarray] = None

    def select(self, x: np.ndarray, y: np.ndarray, options: Optional[Dict] = None) -> np.ndarray:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): options in the optimizations provided
                to scipy.optimize.minimize
        Returns: list of int indices
        """

        n_data, n_dim = x.shape
        options = options or {"maxiter": 1e4, "ftol": 1e-12}
        res = minimize(
            lambda beta: self.construct_loss(x=x, y=y, beta=beta),
            [0] * n_dim,
            jac=self.construct_jac(x=x, y=y),
            method=self.method,
            constraints=self.construct_constraints(x=x, y=y),
            options=options,
        )
        if res.status != 0:
            raise RuntimeError(f"Not converged, status {res.status}")
        self.is_fitted = True
        self.coef_ = res.x
        # output coefficient indices that are above certain thresholds
        self.indices = np.where(np.abs(self.coef_) > self.coef_thres)[0]
        self.coef_[np.where(np.abs(self.coef_) <= self.coef_thres)[0]] = 0.0
        return self.indices

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: np.ndarray) -> float:
        """
        Get loss function from data and tentative coefficients beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: loss value
        """
        raise NotImplementedError

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: Optional[np.ndarray] = None
    ) -> Optional[Union[Dict, List, NonlinearConstraint]]:
        """
        Get constraints dictionary from data, e.g.,
        {"func": lambda beta: fun(x, y, beta), "type": "ineq"}
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): parameter to optimize
        Returns: dict of constraints
        """
        return None

    def construct_jac(self, x: np.ndarray, y: np.ndarray) -> Optional[Callable]:
        """
        Jacobian of cost function
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: Jacobian function
        """
        return None

    def evaluate(self, x: np.ndarray, y: np.ndarray, metric: str = "neg_mean_absolute_error") -> float:
        """
        Evaluate the linear models using x, and y test data
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            metric (str): scorer function, used with
                sklearn.metrics.get_scorer
        Returns:
        """
        metric_func = get_scorer(metric)
        lr = LinearRegression(fit_intercept=False)
        lr.coef_ = self.coef_[self.indices]  # type: ignore
        lr.intercept_ = 0
        return metric_func(lr, x[:, self.indices], y)

    def get_coef(self) -> np.ndarray:
        """
        Get coefficients
        Returns: the coefficients array
        """
        return self.coef_

    def get_feature_indices(self) -> np.ndarray:
        """
        Get selected feature indices
        Returns:
        """
        return self.indices

    def predict(self, x: np.ndarray) -> np.ndarray:
        """
        Predict the results using sparsified coefficients
        Args:
            x (np.ndarray): design matrix
        Returns:
        """
        return x[:, self.indices].dot(self.coef_[self.indices])  # type: ignore

    def compute_residual(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Compute
        Args:
            x (np.ndarray): design matrix
            y (np.ndarray): target vector
        Returns: residual vector
        """
        return y - self.predict(x)

    @classmethod
    def _get_param_names(cls):
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            return []
        init_signature = inspect.signature(init)
        parameters = [p for p in init_signature.parameters.values() if p.name != "self" and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_KEYWORD:
                raise RuntimeError(
                    "scikit-learn estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    " %s with constructor %s doesn't "
                    " follow this convention." % (cls, init_signature)
                )
        return sorted([p.name for p in parameters])

    def get_params(self):
        """
        Get params for this selector

        Returns: mapping of string to any
                parameter names mapped to their values

        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            out[key] = value
        return out

    def set_params(self, **params):
        """
        Set the parameters of this selector
        Args:
            **params: dict
            Selector parametrs

        Returns:
            self: selector instance

        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                raise ValueError(
                    "Invalid parameter %s for selector %s. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`." % (key, self)
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self


class DantzigSelector(BaseSelector):
    """
    Equation 11 in
    https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf
    and reference in https://projecteuclid.org/download/pdfview_1/euclid.aos/1201012958
    """

    def __init__(self, lambd, sigma=1.0, **kwargs):
        """
        Dantzig selector
        Args:
            lamb: tunable parameter
            sigma: standard deviation of the error
        """
        self.lambd = lambd
        self.sigma = sigma
        super().__init__(**kwargs)

    def construct_loss(self, x, y, beta) -> float:
        """
        Get loss function from data and tentative coefficients beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: loss value
        """
        return np.sum(np.abs(beta)).item()

    def construct_jac(self, x: np.ndarray, y: np.ndarray) -> Callable:
        """
        Jacobian of cost functions
        Args:
            x:
            y:
        Returns:
        """

        def _jac(beta):
            sign = np.sign(beta)
            sign[np.abs(sign) < 0.1] = 1.0
            sign *= 30.0  # multiply the gradients to get better convergence
            return sign

        return _jac

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: Optional[np.ndarray] = None
    ) -> NonlinearConstraint:
        """
        Get constraints dictionary from data, e.g.,
        {"func": lambda beta: fun(x, y, beta), "type": "ineq"}
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): placeholder
        Returns: dict of constraints
        """

        def _constraint(beta):
            return np.linalg.norm(x.T @ (y - x @ beta), np.infty)

        def _jac(beta):
            vec = x.T @ (y - x @ beta)
            max_ind = np.argmax(np.abs(vec))
            der = np.zeros_like(vec.ravel())
            der[max_ind] = np.sign(vec[max_ind])
            return -x.T.dot(x).dot(der)

        return NonlinearConstraint(_constraint, -np.infty, self.lambd * self.sigma, jac=_jac)


class PenalizedLeastSquares(BaseSelector):
    """
    Penalized least squares. In addition to minimizing the sum of squares loss,
    it adds an additional penalty to the coefficients
    """

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: np.ndarray) -> float:
        """
        Construct the loss function. An extra penalty term is added
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: sum of errors
        """
        n = x.shape[0]
        se = 1.0 / (2 * n) * np.sum((y - x.dot(beta)) ** 2) + self.penalty(beta, x=x, y=y)
        return se

    def _sse_jac(self, x, y, beta):
        n = x.shape[0]
        return 1.0 / n * (y - x.dot(beta)).T.dot(-x)

    def _penalty_jac(self, x, y, beta):
        return 0.0

    def construct_jac(self, x: np.ndarray, y: np.ndarray):
        """
        Construct the jacobian of loss function
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: jacobian vector
        """

        def _jac(beta):
            return self._sse_jac(x, y, beta) + self._penalty_jac(x, y, beta)

        return _jac

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: Optional[np.ndarray] = None
    ) -> List[Optional[Dict]]:
        """
        No constraints
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): placeholder only
        Returns: a list of dictionary constraints
        """
        return []

    def penalty(self, beta: np.ndarray, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None) -> float:
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value
        """
        return 0.0


class SCAD(PenalizedLeastSquares):
    """
    Smoothly clipped absolute deviation (SCAD),
    equation 12 and 13 in https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf
    """

    def __init__(self, lambd: Union[float, np.ndarray], a: float = 3.7, **kwargs):
        """
        Smoothly clipped absolute deviation.
        Args:
            lambd (float or list of floats): The weights for the penalty
            a (float): hyperparameter in SCAD penalty
        """
        self.lambd = lambd
        self.a = a
        super().__init__(**kwargs)

    def penalty(self, beta: np.ndarray, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None) -> float:
        """
        Calculate the SCAD penalty from input x, output y
            and coefficient beta
        Args:
            beta (np.ndarray): N coefficients
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: penalty value
        """
        beta_abs = np.abs(beta)
        penalty = (
            self.lambd * beta_abs * (beta_abs <= self.lambd)
            + -(beta_abs ** 2 - 2 * self.a * self.lambd * beta_abs + self.lambd ** 2)
            / (2 * (self.a - 1))
            * (beta_abs > self.lambd)
            * (beta_abs <= self.a * self.lambd)
            + (self.a + 1) * self.lambd ** 2 / 2.0 * (beta_abs > self.a * self.lambd)
        )
        return np.sum(penalty).item()

    def _penalty_jac(self, x, y, beta):
        beta = np.abs(beta)
        z = self.a * self.lambd - beta
        z[z < 0] = 0
        return self.lambd * (beta <= self.lambd + z / ((self.a - 1) * self.lambd) * (beta > self.lambd))


class Lasso(PenalizedLeastSquares):
    """
    Simple Lasso regression
    """

    def __init__(self, lambd, **kwargs):
        """
        Lasso regression with lambda * norm_1(beta) as penalty
        Args:
            lambd (float): weights for the penalty
            **kwargs:
        """
        self.lambd = lambd
        super().__init__(**kwargs)

    def penalty(self, beta: np.ndarray, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None) -> float:
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            beta (np.ndarray): N coefficients
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: penalty value
        """
        beta_abs = np.abs(beta)
        return np.sum(self.lambd * beta_abs).item()

    def _penalty_jac(self, x, y, beta):
        sign = np.sign(beta)
        sign[np.abs(sign) < 0.2] = 1
        return self.lambd * sign


class AdaptiveLasso(PenalizedLeastSquares):
    """
    Adaptive lasso regression using OLS coefficients
    as the root-n estimator coefficients
    """

    def __init__(self, lambd, gamma, **kwargs):
        """
        Adaptive lasso regression
        Args:
            lambd (float or list of floats):
            gamma (float): exponential for hat(beta)
            **kwargs:
        """
        self.lambd = lambd
        self.gamma = gamma
        self.w = 1
        super().__init__(**kwargs)

    def select(self, x, y, options=None) -> np.ndarray:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): options in the optimizations provided
                to scipy.optimize.minimize
        Returns: list of int indices
        """
        self.w = self.get_w(x, y)
        return super().select(x, y, options)

    def get_w(self, x, y) -> np.ndarray:
        """
        Get adaptive weights from data
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: coefficients array
        """
        beta_hat = lstsq(x, y)[0]
        w = 1.0 / np.abs(beta_hat) ** self.gamma
        return w

    def penalty(self, beta: np.ndarray, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None) -> float:
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            beta (np.ndarray): N coefficients
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: penalty value
        """
        return np.sum(self.lambd * self.w * np.abs(beta)).item()

    def _penalty_jac(self, x, y, beta):
        sign = np.sign(beta)
        sign[np.abs(sign) < 0.2] = 1
        return self.lambd * self.w * sign


class L0BrutalForce(BaseSelector):
    """
    Brutal force combinatorial screening of features.
    This method takes all possible combinations of features
    and optimize the following loss function
        1/2 * mean((y-x @ beta)**2) + lambd * |beta|_0
    """

    def __init__(self, lambd: float, **kwargs):
        """
        Initialization of L0 optimization
        Args:
            lambd (float): penalty term
            **kwargs:
        """
        self.lambd = lambd
        super().__init__(**kwargs)

    def select(self, x: np.ndarray, y: np.ndarray, options: Optional[Dict] = None, n_job: int = 1) -> np.ndarray:
        """
        L0 combinatorial optimization
        Args:
            x (np.ndarray): design matrix
            y (np.ndarray): target vector
            options:
        Returns:
        """
        n, p = x.shape
        index_array = list(range(p))

        def _lstsq(c):
            x_comb = x[:, c]
            beta = lstsq(x_comb, y)[0]
            res = 1.0 / 2 * np.mean((x_comb.dot(beta) - y) ** 2)
            penalty = self.lambd * len(c)
            res += penalty
            return res

        indices = []
        for p_temp in range(1, p + 1):
            for comb in combinations(index_array, p_temp):
                indices.append(comb)
        loss = Parallel(n_jobs=n_job)(delayed(_lstsq)(comb) for comb in indices)
        argmin = np.argmin(loss)
        self.indices = np.array(indices[argmin])
        x_temp = x[:, self.indices]
        self.coef_ = np.zeros_like(x[0, :])
        self.coef_[self.indices] = lstsq(x_temp, y)[0]
        return self.indices



# coding: utf-8
# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""
Define abstract base classes.
"""
from ._data import BaseDataSource  # noqa
from ._describer import BaseDescriber, SequentialDescriber  # noqa
from ._describer import DummyDescriber, describer_type  # noqa
from ._model import BaseModel, KerasModel, SKLModel  # noqa
from ._model import is_keras_model, is_sklearn_model
from ._feature_batch import get_feature_batch
from ._mixin import TargetScalerMixin


__all__ = [
    "BaseDataSource",
    "BaseDescriber",
    "SequentialDescriber",
    "DummyDescriber",
    "BaseModel",
    "KerasModel",
    "SKLModel",
    "get_feature_batch",
    "describer_type",
    "is_keras_model",
    "is_sklearn_model",
    "TargetScalerMixin",
]



"""
Deep learning module
"""

from ._atomsets import AtomSets  # noqa
from ._layers import WeightedAverageLayer, WeightedSet2Set  # noqa
from ._mlp import MLP  # noqa

__all__ = ["WeightedAverageLayer", "WeightedSet2Set", "AtomSets", "MLP"]



# coding: utf-8
# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.
"""
Base classes for potentials
"""

import abc

from maml.base import BaseModel


class PotentialMixin:
    """
    Abstract base class for a Interatomic Potential.
    """

    @abc.abstractmethod
    def train(self, train_structures, train_energies, train_forces, train_stresses, **kwargs):
        """
        Train interatomic potentials with energies, forces ann stresses corresponding
            to structures.

        Args:
            train_structures (list): List of Pymatgen Structure objects.
            train_energies (list): List of DFT-calculated total energies of each
                structure in structures list.
            train_forces (list): List of DFT-calculated (m, 3) forces of each
                structure with m atoms in structures list. m can be varied with
                each single structure case.
            train_stresses (list): List of DFT-calculated (6, ) virial stresses of
                each structure in structures list.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def evaluate(self, test_structures, test_energies, test_forces, test_stresses):
        """
        Evaluate energies, forces and stresses of structures with trained
        interatomic potentials.

        Args:
            test_structures (list): List of Pymatgen Structure Objects.
            test_energies (list): List of DFT-calculated total energies of each
                structure in structures list.
            test_forces (list): List of DFT-calculated (m, 3) forces of each
                structure with m atoms in structures list. m can be varied with
                each single structure case.
            test_stresses (list): List of DFT-calculated (6, ) viriral stresses of
                each structure in structures list.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict_efs(self, structure):
        """
        Predict energy, forces and stresses of the structure.

        Args:
            structure (Structure): Pymatgen Structure object.

        Returns:
            energy, forces, stress
        """
        pass

    @abc.abstractmethod
    def write_param(self):
        """
        Write optimized weights file to perform energy and force prediction via
        lammps calculation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def from_config(self, **kwargs):
        """
        Initialize potentials with parameters file.
        """
        raise NotImplementedError


class Potential(PotentialMixin, BaseModel):  # type: ignore
    """
    Potential models that can be used to fit structure-[energy, force, stress]
    pairs
    """

    pass



"""
Materials Project DataSource
"""
from typing import Dict, List, Union

import pandas as pd
from pymatgen.ext.matproj import MPRester

from maml import BaseDataSource


class MaterialsProject(BaseDataSource):
    """
    Query the Materials Project for Data.
    """

    def __init__(self, api_key: str = None):
        """
        Args:
            api_key (str or None): Materials Project API key.
        """
        self.mpr = MPRester(api_key=api_key)

    def get(self, criteria: Union[str, Dict], properties: List[str]) -> pd.DataFrame:  # type: ignore
        """
        Args:
            criteria (str or dict): Criteria for query
            properties (list): Properties to be queried.

        Returns:
            pandas DataFrame
        """
        data = self.mpr.query(criteria=criteria, properties=properties)
        return pd.DataFrame(data)



"""
MAML models base classes
"""
from typing import Any, Union, List, Optional, Callable

import joblib
import numpy as np

from maml.base import DummyDescriber, BaseDescriber
from maml.utils import get_full_args, to_array


class BaseModel:
    """
    Abstract Base class for a Model. Basically, it usually wraps around a deep
    learning package, e.g., the Sequential Model in Keras, but provides for
    transparent conversion of arbitrary input and outputs.
    """

    def __init__(self, model: Any, describer: Optional[BaseDescriber] = None, **kwargs):
        """
        Args:
            model (Any): ML models, for example, sklearn models or keras models
            describer (BaseDescriber): Describer that converts object into features
        """
        if describer is None:
            describer = DummyDescriber()
        self.describer = describer
        self.model = model

    def fit(
        self,
        features: Union[List, np.ndarray],
        targets: Optional[Union[List, np.ndarray]] = None,
        val_features: Optional[Union[List, np.ndarray]] = None,
        val_targets: Optional[Union[List, np.ndarray]] = None,
        **kwargs,
    ) -> "BaseModel":
        """
        Args:
            features (list or np.ndarray): Numerical input feature list or
                numpy array with dim (m, n) where m is the number of data and
                n is the feature dimension.
            targets (list or np.ndarray): Numerical output target list, or
                numpy array with dim (m, ).
            val_features (list or np.ndarray): validation features
            val_targets (list or np.ndarray): validation targets

        Returns:
            self
        """
        self.model.fit(features, targets, **kwargs)  # type: ignore
        return self

    def train(
        self,
        objs: Union[List, np.ndarray],
        targets: Optional[Union[List, np.ndarray]] = None,
        val_objs: Optional[Union[List, np.ndarray]] = None,
        val_targets: Optional[Union[List, np.ndarray]] = None,
        **kwargs,
    ) -> "BaseModel":
        """
        Train the models from object, target pairs

        Args:
            objs (list of objects): List of objects
            targets (list): list of float or np.ndarray
            val_objs (list of objects): list of validation objects
            val_targets (list): list of validation targets
            **kwargs:

        Returns: self

        """
        features = to_array(self.describer.fit_transform(objs))
        targets = to_array(targets)

        if (val_objs is None) and (val_targets is not None):
            raise ValueError("training objects are none, but the targets are not")
        if val_objs is not None:
            val_features = to_array(self.describer.transform(val_objs))
            val_targets = to_array(val_targets)
        else:
            val_features = None
            val_targets = None
        return self.fit(
            features=features, targets=targets, val_features=val_features, val_targets=val_targets, **kwargs
        )

    def _predict(self, features: np.ndarray, **kwargs) -> np.ndarray:
        """
        Predict the values given a set of inputs based on fitted models.

        Args:
            features (np.ndarray): array-like input features.

        Returns:
            List of output objects.
        """
        return self.model.predict(features, **kwargs)  # type: ignore

    def predict_objs(self, objs: Union[List, np.ndarray]) -> np.ndarray:
        """
        Predict the values given a set of objects. Usually Pymatgen
            Structure objects.
        """
        return self._predict(self.describer.transform(objs))


class SklearnMixin:
    """
    Sklearn models save and load functionality
    """

    def save(self, filename: str):
        """Save the models and describers

        Arguments:
            filename (str): filename for save
        """
        joblib.dump({"models": self.model, "describers": self.describer}, filename)

    def load(self, filename: str):
        """
        Load models parameters from filename
        Args:
            filename (str): models file name

        Returns: None

        """
        m = joblib.load(filename)
        self.model = m["models"]
        self.describer = m["describers"]

    @classmethod
    def from_file(cls, filename: str, **kwargs):
        """
        Load the models from file
        Args:
            filename (str): filename
            **kwargs:

        Returns: object instance

        """
        instance = cls(model=None, **kwargs)  # type: ignore
        instance.load(filename)
        return instance

    def evaluate(
        self,
        eval_objs: Union[List, np.ndarray],
        eval_targets: Union[List, np.ndarray],
        is_feature: bool = False,
        metric: Union[str, Callable] = None,
    ) -> np.ndarray:
        """
        Evaluate objs, targets

        Args:
            eval_objs (list): objs for evaluation
            eval_targets (list): target list for the corresponding objects
            is_feature (bool): whether the input x is feature matrix
            metric (callable): metric for evaluation
        """

        from sklearn.metrics import check_scoring

        eval_features = eval_objs if is_feature else self.describer.transform(eval_objs)
        scorer = check_scoring(self.model, scoring=metric)
        return scorer(self.model, eval_features, eval_targets)


class KerasMixin:
    """
    keras models mixin with save and load functionality
    """

    def save(self, filename: str):
        """Save the models and describers

        Arguments:
            filename (str): filename for save
        """
        joblib.dump(self.describer, filename)
        self.model.save(filename + ".hdf5")

    def load(self, filename: str, custom_objects: List = None):
        """
        Load models parameters from filename
        Args:
            filename (str): models file name

        Returns: None

        """
        import tensorflow as tf

        self.describer = joblib.load(filename)
        self.model = tf.keras.models.load_model(filename + ".hdf5", custom_objects=custom_objects)

    @classmethod
    def from_file(cls, filename: str, **kwargs):
        """
        Load the models from file
        Args:
            filename (str): filename
            **kwargs:

        Returns: object instance

        """
        instance = cls(model=None, **kwargs)  # type: ignore
        instance.load(filename)
        return instance

    def evaluate(
        self, eval_objs: Union[List, np.ndarray], eval_targets: Union[List, np.ndarray], is_feature: bool = False
    ) -> np.ndarray:
        """
        Evaluate objs, targets

        Args:
            eval_objs (list): objs for evaluation
            eval_targets (list): target list for the corresponding objects
            is_feature (bool): whether the input x is feature matrix
            metric (callable): metric for evaluation
        """

        eval_features = eval_objs if is_feature else self.describer.transform(eval_objs)
        return self.model.evaluate(to_array(eval_features), to_array(eval_targets))  # type: ignore

    @staticmethod
    def get_input_dim(describer: Optional[BaseDescriber] = None, input_dim: Optional[int] = None) -> Union[int, None]:
        """
        Get feature dimension/input_dim from describers or input_dim

        Args:
            describer (Describer): describers
            input_dim (int): optional input dim int
        """
        if input_dim is not None:
            feature_size: Union[int, None] = input_dim
        elif describer is not None:
            feature_size = describer.feature_dim
        else:
            feature_size = None
        return feature_size


class SKLModel(BaseModel, SklearnMixin):
    """MAML models with sklearn models as estimator"""

    def __init__(self, model: Any, describer: Optional[BaseDescriber] = None, **kwargs):
        """
        Args:
            model (Any): ML models, for example, sklearn models or keras models
            describer (BaseDescriber): Describer that converts object into features
        """
        super().__init__(model=model, describer=describer, **kwargs)


class KerasModel(BaseModel, KerasMixin):
    """MAML models with keras models as estimators"""

    def __init__(self, model: Any, describer: Optional[BaseDescriber] = None, **kwargs):
        """
        Args:
            model (Any): ML models, for example, sklearn models or keras models
            describer (BaseDescriber): Describer that converts object into features
        """
        super().__init__(model=model, describer=describer, **kwargs)

    def fit(
        self,
        features: Union[List, np.ndarray],
        targets: Optional[Union[List, np.ndarray]] = None,
        val_features: Optional[Union[List, np.ndarray]] = None,
        val_targets: Optional[Union[List, np.ndarray]] = None,
        **kwargs,
    ) -> "BaseModel":
        """
        Args:
            features (list or np.ndarray): Numerical input feature list or
                numpy array with dim (m, n) where m is the number of data and
                n is the feature dimension.
            targets (list or np.ndarray): Numerical output target list, or
                numpy array with dim (m, ).
            val_features (list or np.ndarray): validation features
            val_targets (list or np.ndarray): validation targets

        Returns:
            self
        """
        if (val_features is None) and (val_targets is not None):
            raise ValueError("Validation data incorrect")
        targets = np.array(targets)
        val_targets = np.array(val_targets)
        fit_args = get_full_args(self.model.fit)  # type: ignore

        # construct validation data for keras models
        val_kwargs = {}
        kw_remove = []
        for kw in kwargs:
            if kw not in fit_args:
                val_kwargs.update({kw: kwargs.get(kw)})
                kw_remove.append(kw)

        for kw in kw_remove:
            kwargs.pop(kw)

        if val_features is not None:
            validation_data = self._get_validation_data(val_features, val_targets, **val_kwargs)
        else:
            validation_data = None
        self.model.fit(features, targets, validation_data=validation_data, **kwargs)  # type: ignore
        return self

    @staticmethod
    def _get_validation_data(val_features, val_targets, **val_kwargs):
        """
        construct validation data, the default is just returning a list of
        val_features and val_targets
        """
        return val_features, val_targets


def is_sklearn_model(model: BaseModel) -> bool:
    """
    Check whether the model is sklearn
    Args:
        model (BaseModel): model
    Returns: bool
    """
    from sklearn.base import BaseEstimator

    return isinstance(model.model, BaseEstimator)


def is_keras_model(model: BaseModel) -> bool:
    """
    Check whether the model is keras
    Args:
        model (BaseModel): model
    Returns: bool
    """
    from tensorflow.keras.models import Model

    return isinstance(model.model, Model)



"""
Class for performing local environment predictions
"""

import json
import logging
import os
import warnings
from numbers import Number
from typing import List, Optional, Union
from zipfile import ZipFile
import requests
import numpy as np

import joblib

from scipy.interpolate import interp1d
from tqdm import tqdm

from maml.apps.rfxas._core import XANES


logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

CWD = os.path.dirname(os.path.abspath(__file__))

DATA_DIR = os.path.join(CWD, "data")
EXTRA_MODELS = os.path.join(CWD, "ex_models.zip")
EXTRA_MODEL_DIR = os.path.join(CWD, "ex_models/Model_share")
EXTRA_MODEL_SUB_DICT = {"cnn": "CNN_model", "knn": "KNN_model", "mlp": "MLP_model", "svc": "SVC_model"}
EXTRA_TEMPLATES = {"cnum": "{}_c_num", "cmotif": "{}_c_env_{}_env_label"}

EXTRA_MODEL_URL = "https://ndownloader.figshare.com/files/18741755"
RF_MODEL_PATH = os.path.join(CWD, "models.zip")
RF_MODEL_URL = "https://ndownloader.figshare.com/files/25946756"


CN_CLASSES = [
    "CN_1-CN_4",
    "CN_12",
    "CN_12-CN_10",
    "CN_12-CN_3",
    "CN_12-CN_6-CN_9",
    "CN_2",
    "CN_2-CN_4-CN_6",
    "CN_2-CN_6",
    "CN_2-CN_6-CN_5",
    "CN_3",
    "CN_3-CN_4",
    "CN_3-CN_5",
    "CN_3-CN_5-CN_6",
    "CN_3-CN_6",
    "CN_3-CN_6-CN_4",
    "CN_4",
    "CN_4-CN_1",
    "CN_4-CN_2",
    "CN_4-CN_3",
    "CN_4-CN_3-CN_5-CN_6",
    "CN_4-CN_3-CN_6",
    "CN_4-CN_5",
    "CN_4-CN_5-CN_6",
    "CN_4-CN_6",
    "CN_4-CN_6-CN_5",
    "CN_4-CN_8",
    "CN_5",
    "CN_5-CN_3",
    "CN_5-CN_3-CN_4",
    "CN_5-CN_3-CN_6",
    "CN_5-CN_4",
    "CN_5-CN_4-CN_6",
    "CN_5-CN_6",
    "CN_5-CN_6-CN_3",
    "CN_5-CN_6-CN_7",
    "CN_5-CN_7-CN_6",
    "CN_6",
    "CN_6-CN_1",
    "CN_6-CN_12",
    "CN_6-CN_2",
    "CN_6-CN_3",
    "CN_6-CN_3-CN_5",
    "CN_6-CN_4",
    "CN_6-CN_4-CN_5",
    "CN_6-CN_5",
    "CN_6-CN_5-CN_3",
    "CN_6-CN_7",
    "CN_6-CN_8",
    "CN_6-CN_8-CN_4",
    "CN_7",
    "CN_7-CN_6",
    "CN_7-CN_9",
    "CN_8",
    "CN_8-CN_7",
]


with open(os.path.join(DATA_DIR, "cmotif_labels.json"), "r") as f:
    CMOTIF_LABELS = json.load(f)


class CenvPrediction:
    """
    Coordination environment predictor. Supported models are
        RandomForest (rf)
        K-neareset Neighbors (knn)
        Support-vector Classifier (svc)
        Multi-layer Perceptron (mlp)
        Convolutional Neural Network (cnn)
    """

    _available_models = ["rf", "knn", "svc", "mlp", "cnn"]

    def __init__(
        self,
        xanes_spectrum: XANES,
        energy_reference: str,
        energy_range: Union[float, List[float]],
        edge_energy: Optional[float] = None,
        spectrum_interpolation: bool = True,
        model: str = "rf",
        model_dir: Optional[str] = None,
        extra_model_dir: str = EXTRA_MODEL_DIR,
    ):
        """
        Args:
            xanes_spectrum: XANES object
            energy_reference (str): Energy reference mode, choose from 'lowest' or 'E0'. 'lowest' mode for
                            using the lowest energy of spectrum as the starting point of the to be characterized
                            spectrum. 'E0' mode for using the edge energy as the reference point to generate to be
                            characterized spectrum energy range.
            energy_range (list/float): Energy range of spectrum used for prediction. If the energy reference mode is
                'lowest', energy range value need to be a number specifies the energy range. If the energy
                reference mode is 'E0', energy range ought to be a list of two numbers. The first number
                (negative) represents the difference between the lower bound energy and energy reference.
            edge_energy (float): Edge energy of spectra. Usually determined using MBACK algorithm and provided by users
            spectrum_interpolation: Whether or not do spectrum interpolation. Default to True. If
                spectrum_interpolation option if False, then the CenvPrediction object will use the original spectrum
                of xanes_spectrum for coordination environment prediction. The original spectrum need to be a vector
                with length equals 200.
            model (str): rf, knn, mlp or svc. Default is rf. For other models, models download ~1.28 GB is required.
            model_dir (str): specifies where the models random forest models are stored.
                if the dir does not exists, the models will be downloaded from figshare
            extra_model_dir (str): specifies where the extra models are stored.
                if the dir does not exists, the extra models will be downloaded from figshare
        """

        self.xanes_spectrum = xanes_spectrum
        self.absorption_specie = self.xanes_spectrum.absorption_specie
        self.energy_reference = energy_reference
        self.energy_range = energy_range
        self.cnum_model_name_template = "RandomForest_{}_c_num.sav"
        self.cmotif_model_name_template = "RandomForest_{}_c_env_ex_{}.sav"
        if model_dir is None:
            model_dir = os.path.join(CWD, "models")
        self.model_dir = model_dir

        if not os.path.isdir(self.model_dir):
            _download_models(RF_MODEL_URL, RF_MODEL_PATH, self.model_dir)

        if model not in self._available_models:
            raise ValueError("Model type %s not recognized " % str(model), " choose from ", self._available_models)
        if model != "rf":
            if os.path.isdir(extra_model_dir):
                pass
            else:
                _download_models(EXTRA_MODEL_URL, EXTRA_MODELS, extra_model_dir)

            self.model_dir = os.path.join(extra_model_dir, EXTRA_MODEL_SUB_DICT[model])

            if model.lower() == "cnn":
                suffix = ".h5"
            else:
                suffix = ".sav"
            self.cnum_model_name_template = model.upper() + "_" + EXTRA_TEMPLATES["cnum"] + suffix
            self.cmotif_model_name_template = model.upper() + "_" + EXTRA_TEMPLATES["cmotif"] + suffix

        if isinstance(self.energy_range, list):
            self.energy_lower_bound = self.energy_range[0]
            self.energy_higher_bound = self.energy_range[-1]

        self._parameter_validation()

        if energy_reference == "E0" and edge_energy is None:
            warnings.warn(
                "Using edge energy of xanes_spectrum object, be cautious about how the object's edge energy"
                " is determined"
            )
            self.edge_energy = self.xanes_spectrum.e0
        elif energy_reference == "E0" and edge_energy:
            self.edge_energy = edge_energy
        elif self.energy_reference == "lowest":
            self.edge_energy = self.xanes_spectrum.e0

        if spectrum_interpolation:
            self._energy_interp()
        else:
            self.interp_spectrum = self.xanes_spectrum.y
            self.interp_spectrum_reshape = np.array(self.interp_spectrum).reshape(1, -1)
            self.interp_energy = self.xanes_spectrum.x

        self.model = model.lower()
        if self.model == "cnn":
            self.interp_spectrum_reshape = self.interp_spectrum_reshape.reshape((1, -1, 1))

    def cenv_prediction(self):
        """
        Function used for coordination environment prediction

        Returns:
            Two new attributes of CenvPrediction object. The attribute 'pred_cnum_ranklist' is the predicted
            coordination number ranklist. The attribute pred_cenv contains the final predicted coordination
            environment ranklists.

        """
        self._cnum_prediction()
        self._cmotif_prediction()

    def _cnum_prediction(self):
        cnum_pred_ele_json = os.path.join(DATA_DIR, "cnum_predict_elements.json")
        with open(cnum_pred_ele_json, "r") as fp:
            cnum_pred_elements = json.load(fp)

        if self.absorption_specie not in cnum_pred_elements:
            warning_msg = "Coordination number prediction models for {} is unavailable.".format(self.absorption_specie)
            warnings.warn(warning_msg)
            self.pred_cnum_ranklist = "cnum undetermined"
        else:
            cnum_model_name = self.cnum_model_name_template.format(self.absorption_specie)
            cnum_model_path = os.path.join(self.model_dir, "cnum", cnum_model_name)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                logger.info("Loaded %s models" % cnum_model_path)
                if cnum_model_path.endswith(".sav"):
                    cnum_model_loaded = joblib.load(cnum_model_path)
                elif cnum_model_path.endswith(".h5"):
                    import tensorflow as tf

                    print("cnum ", cnum_model_path)
                    cnum_model_loaded = tf.keras.models.load_model(cnum_model_path)
                else:
                    raise ValueError("Cnum models not recognized")
            self.pred_cnum_ranklist = cnum_model_loaded.predict(self.interp_spectrum_reshape)
            if self.model == "cnn":
                self.pred_cnum_ranklist = CN_CLASSES[np.argmax(self.pred_cnum_ranklist)]
            else:
                self.pred_cnum_ranklist = self.pred_cnum_ranklist[0]

    def _cmotif_prediction(self):
        cmotif_pred_ele_json = os.path.join(DATA_DIR, "cmotif_predict_elements.json")
        with open(cmotif_pred_ele_json, "r") as fp:
            cmotif_ele_env_dict = json.load(fp)
        ele_env_valid_prediction = cmotif_ele_env_dict[self.absorption_specie]

        if self.pred_cnum_ranklist == "cnum undetermined":
            self.pred_cenv = "cenv undetermined"
        else:
            pred_cnum_ranklist = self.pred_cnum_ranklist
            # Using predicted cnum ranklist to predict cenv ranklist
            spectral_env_pred = []
            for indi_pred_cnum in pred_cnum_ranklist.split("-"):
                cmotif_pred_cenv = "ex_{}".format(indi_pred_cnum)

                # No available coord. motif prediction models for this particular coord. num. of this element
                if cmotif_pred_cenv not in ele_env_valid_prediction:
                    pseudo_cmotif_label = "{} coord. motif undetermined".format(indi_pred_cnum)
                    spectral_env_pred.append(pseudo_cmotif_label)
                else:
                    cmotif_model_name = self.cmotif_model_name_template.format(self.absorption_specie, indi_pred_cnum)
                    cmotif_model_path = os.path.join(self.model_dir, "cmotif", cmotif_model_name)
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=UserWarning)
                        logger.info("Loaded %s models" % cmotif_model_path)
                        if cmotif_model_path.endswith(".sav"):
                            cmotif_model_loaded = joblib.load(cmotif_model_path)
                        elif cmotif_model_path.endswith(".h5"):
                            import tensorflow as tf

                            print("cmotif", cmotif_model_path)
                            cmotif_model_loaded = tf.keras.models.load_model(cmotif_model_path)
                        else:
                            raise ValueError("Model format not recognized")

                        if self.model == "cnn":
                            pred_motif_ranklist = cmotif_model_loaded.predict(
                                self.interp_spectrum_reshape.reshape((1, -1, 1))
                            )
                            labels = CMOTIF_LABELS[cmotif_model_name[4:-3]]
                            pred_motif_ranklist = labels[np.argmax(pred_motif_ranklist)]
                        else:
                            pred_motif_ranklist = cmotif_model_loaded.predict(self.interp_spectrum_reshape)[0]
                        pred_cnum_cmotif_concat = "-".join([indi_pred_cnum, pred_motif_ranklist])
                        spectral_env_pred.append(pred_cnum_cmotif_concat)

            self.pred_cenv = spectral_env_pred

    def _energy_interp(self):
        # if energy_reference is 'lowest' and energy range is proper passed
        if self.energy_reference == "lowest":
            x_axis_energy = self.xanes_spectrum.x
            y_spectrum = self.xanes_spectrum.y
            x_axis_energy_start = np.min(x_axis_energy)
            x_axis_energy_end = x_axis_energy_start + self.energy_range
            x_axis_energy_end_index = find_nearest_energy_index(x_axis_energy, x_axis_energy_end)
            x_axis_linspace = np.linspace(x_axis_energy_start, x_axis_energy[x_axis_energy_end_index], num=200)
            f = interp1d(x_axis_energy, y_spectrum, kind="cubic", bounds_error=False, fill_value=0)
            self.interp_spectrum = f(x_axis_linspace)
            normalize_factor = np.max(self.interp_spectrum, axis=0)
            self.interp_spectrum /= normalize_factor
            self.interp_energy = x_axis_linspace
            self.interp_spectrum_reshape = np.array(self.interp_spectrum).reshape(1, -1)

        elif self.energy_reference == "E0":
            x_axis_energy = self.xanes_spectrum.x
            y_spectrum = self.xanes_spectrum.y
            x_axis_energy_start = self.edge_energy + self.energy_lower_bound
            x_energy_start_index = find_nearest_energy_index(x_axis_energy, x_axis_energy_start)
            x_axis_energy_end = self.edge_energy + self.energy_higher_bound
            x_energy_end_index = find_nearest_energy_index(x_axis_energy, x_axis_energy_end)
            x_axis_linspace = np.linspace(
                x_axis_energy[x_energy_start_index], x_axis_energy[x_energy_end_index], num=200
            )
            f = interp1d(x_axis_energy, y_spectrum, kind="cubic", bounds_error=False, fill_value=0)
            self.interp_spectrum = f(x_axis_linspace)
            normalize_factor = np.max(self.interp_spectrum, axis=0)
            self.interp_spectrum /= normalize_factor
            self.interp_spectrum_reshape = np.array(self.interp_spectrum).reshape(1, -1)
            self.interp_energy = x_axis_linspace

    def _parameter_validation(self):
        if self.energy_reference not in ["lowest", "E0"]:
            raise ValueError('Invalid energy reference option, energy_reference should either be "lowest" or "E0"')

        if self.energy_reference == "lowest":
            if not isinstance(self.energy_range, Number):
                raise ValueError(
                    "Energy range needs to be a number when the energy reference point is the starting energy of "
                    "the spectrum"
                )

            if self.energy_range < 0.0:  # type: ignore
                raise ValueError("Energy range needs to be larger than 0. Invalid energy range error.")

        if self.energy_reference == "E0":
            if not isinstance(self.energy_range, list):
                raise ValueError(
                    "Energy range needs to be a list contains lower energy bound and higher energy bound refer to "
                    "energy reference point"
                )

            if self.energy_lower_bound > 0:
                raise ValueError("Energy lower bound needs to be less than zero.")

            if self.energy_higher_bound < 0:
                raise ValueError("Energy higher bound needs to be larger than zero.")


def find_nearest_energy_index(energy_array, energy_value):
    """
    Given a target energy value, returns a value index of energy_array with index value most close to the target energy
    value

    Args:
        energy_array: Energy array to search for the closest energy value index
        energy_value: Target energy value for index searching.

    Returns:

    """
    energy_array = np.asarray(energy_array)
    energy_index = (np.abs(energy_array - energy_value)).argmin()
    return energy_index


def _download_models(url, file_path=EXTRA_MODELS, dest="ext_models"):
    """
    Download machine learning models files

    Args:
        url: (str) url link for the models
        dest: (str) destination for extraction
    """

    logger.info("Fetching {} from {} to {}".format(os.path.basename(file_path), url, file_path))
    if not os.path.isfile(file_path):
        r = requests.get(url, stream=True)
        total_size = int(r.headers.get("content-length", 0))
        block_size = 1024 * 1024  # 1 Mbyte
        t = tqdm(total=total_size, unit="iB", unit_scale=True)
        with open(file_path, "wb") as file_out:
            for chunk in r.iter_content(chunk_size=block_size):
                t.update(len(chunk))
                file_out.write(chunk)
        t.close()
        r.close()
    logger.info("Start extracting models to %s ..." % dest)
    with ZipFile(file_path, "r") as zip_obj:
        zip_obj.extractall(dest)



"""
Materials utils
"""
from typing import Union

from pymatgen.core import Composition, Structure, Molecule


def to_composition(obj: Union[Composition, Molecule, Structure, str]) -> Composition:
    """
    Convert str/structure or composition to compositions

    Args:
        obj (str/structure/composition): object to convert

    Returns:
        Composition object
    """
    if isinstance(obj, str):
        return Composition(obj)
    if isinstance(obj, (Structure, Molecule)):
        return obj.composition
    return obj



# coding: utf-8
# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides basic LAMMPS calculator classes."""

import abc
import io
import itertools
import logging
import os
import subprocess

import numpy as np
from monty.os.path import which
from monty.tempfile import ScratchDir
from pymatgen.core import Structure, Lattice, Element
from pymatgen.io.lammps.data import LammpsData

from maml.apps.pes._base import Potential
from maml.utils import get_lammps_lattice_and_rotation, stress_list_to_matrix, stress_matrix_to_list

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_default_lmp_exe():
    """
    Get lammps executable
    Returns: Lammps executable name
    """

    for lmp_exe in ["lmp_serial", "lmp_mpi"]:
        if which(lmp_exe) is not None:
            logger.info("Setting Lammps executable to %s" % lmp_exe)
            return lmp_exe
    return None


def _pretty_input(lines):
    def prettify(l):
        return (
            l.split()[0].ljust(width) + " ".join(l.split()[1:])
            if not (len(l.split()) == 0 or l.strip().startswith("#"))
            else l
        )

    clean_lines = [l.strip("\n") for l in lines]
    commands = [l for l in clean_lines if len(l.strip()) > 0]
    keys = [c.split()[0] for c in commands if not c.split()[0].startswith("#")]
    width = max([len(k) for k in keys]) + 4
    new_lines = map(prettify, clean_lines)
    return "\n".join(new_lines)


def _read_dump(file_name, dtype="float_"):
    with open(file_name) as f:
        lines = f.readlines()[9:]
    return np.loadtxt(io.StringIO("".join(lines)), dtype=dtype)


class LMPStaticCalculator:
    """
    Abstract class to perform static structure property calculation
    using LAMMPS.
    """

    _COMMON_CMDS = ["units metal", "atom_style charge", "box tilt large", "read_data data.static", "run 0"]

    allowed_kwargs = ["lmp_exe"]

    def __init__(self, **kwargs):
        """
        Initializer for lammps calculator
        Allowed keyword args are lmp_exe string

        """
        lmp_exe = kwargs.pop("lmp_exe", None)
        if lmp_exe is None:
            lmp_exe = get_default_lmp_exe()
        if not which(lmp_exe):
            raise ValueError("lammps executable %s not found" % str(lmp_exe))
        self.LMP_EXE = lmp_exe
        for i, j in kwargs.items():
            if i not in self.allowed_kwargs:
                raise TypeError("%s not in supported kwargs %s" % (str(i), str(self.allowed_kwargs)))
            setattr(self, i, j)

    @abc.abstractmethod
    def _setup(self):
        """
        Setup a calculation, writing input files, etc.

        """
        return

    @abc.abstractmethod
    def _sanity_check(self, structure):
        """
        Check if the structure is valid for this calculation.

        """
        return

    @abc.abstractmethod
    def _parse(self):
        """
        Parse results from dump files.

        """
        return

    def calculate(self, structures):
        """
        Perform the calculation on a series of structures.

        Args:
            structures [Structure]: Input structures in a list.

        Returns:
            List of computed data corresponding to each structure,
            varies with different subclasses.

        """
        for struct in structures:
            assert self._sanity_check(struct) is True, "Incompatible structure found"
        ff_elements = None
        if hasattr(self, "element_profile"):
            element_profile = getattr(self, "element_profile")
            ff_elements = element_profile.keys()
        if hasattr(self, "ff_settings"):
            ff_settings = getattr(self, "ff_settings")
            if hasattr(ff_settings, "elements"):
                ff_elements = getattr(ff_settings, "elements")

        with ScratchDir("."):
            input_file = self._setup()
            data = []
            for struct in structures:
                ld = LammpsData.from_structure(struct, ff_elements)
                ld.write_file("data.static")
                p = subprocess.Popen([self.LMP_EXE, "-in", input_file], stdout=subprocess.PIPE)
                stdout = p.communicate()[0]
                rc = p.returncode
                if rc != 0:
                    error_msg = "LAMMPS exited with return code %d" % rc
                    msg = stdout.decode("utf-8").split("\n")[:-1]
                    try:
                        error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                        error_msg += ", ".join(msg[error_line:])
                    except Exception:
                        error_msg += msg[-1]
                    raise RuntimeError(error_msg)
                results = self._parse()
                data.append(results)
        return data

    def set_lmp_exe(self, lmp_exe: str) -> None:
        """
        Set lammps executable for the instance
        Args:
            lmp_exe (str): lammps executable path
        Returns:
        """
        self.LMP_EXE = lmp_exe


class EnergyForceStress(LMPStaticCalculator):
    """
    Calculate energy, forces and virial stress of structures.
    """

    def __init__(self, ff_settings, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for LAMMPS
                calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
        """
        self.ff_settings = ff_settings
        super().__init__(**kwargs)

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "efs")
        with open(os.path.join(template_dir, "in.efs"), "r") as f:
            input_template = f.read()

        input_file = "in.efs"

        if isinstance(self.ff_settings, Potential):
            ff_settings = self.ff_settings.write_param()
        else:
            ff_settings = self.ff_settings

        with open(input_file, "w") as f:
            f.write(input_template.format(ff_settings="\n".join(ff_settings)))
        return input_file

    def _sanity_check(self, structure):
        return True

    def calculate(self, structures):
        """
        Calculate the energy, forces and stresses of structures.
        Proper rotation of the results are applied when the structure
        is triclinic.

        Args:
            structures (list): a list of structures

        Returns: list of (energy, forces, stresses) tuple

        """
        results = super().calculate(structures=structures)
        final_results = []
        for res, s in zip(results, structures):
            new_forces, new_stresses = self._rotate_force_stress(s, res[1], res[2])
            final_results.append((res[0], new_forces, new_stresses))
        return final_results

    @staticmethod
    def _rotate_force_stress(structure, forces, stresses):
        _, symmop, rot_matrix = get_lammps_lattice_and_rotation(structure)
        inv_rot_matrix = np.linalg.inv(rot_matrix)
        forces = forces.dot(inv_rot_matrix.T)
        stresses = stress_list_to_matrix(stresses, stress_format="LAMMPS")
        stresses = inv_rot_matrix.dot(stresses.dot(inv_rot_matrix.T))
        stresses = stress_matrix_to_list(stresses, stress_format="LAMMPS")
        return forces, stresses

    def _parse(self):
        energy = float(np.loadtxt("energy.txt"))
        force = _read_dump("force.dump")
        stress = np.loadtxt("stress.txt")
        return energy, force, stress


class SpectralNeighborAnalysis(LMPStaticCalculator):
    """
    Calculator for bispectrum components to characterize the local
    neighborhood of each atom in a general way.

    Usage:
        [(b, db, e)] = sna.calculate([Structure])
        b: 2d NumPy array with shape (N, n_bs) containing bispectrum
            coefficients, where N is the No. of atoms in structure and
            n_bs is the No. of bispectrum components.
        db: 2d NumPy array with shape (N, 3 * n_bs * n_elements)
            containing the first order derivatives of bispectrum
            coefficients with respect to atomic coordinates,
            where n_elements is the No. of elements in element_profile.
        e: 2d NumPy array with shape (N, 1) containing the element of
            each atom.

    """

    _CMDS = [
        "pair_style lj/cut 10",
        "pair_coeff * * 1 1",
        "compute sna all sna/atom ",
        "compute snad all snad/atom ",
        "compute snav all snav/atom ",
        "dump 1 all custom 1 dump.element element",
        "dump 2 all custom 1 dump.sna c_sna[*]",
        "dump 3 all custom 1 dump.snad c_snad[*]",
        "dump 4 all custom 1 dump.snav c_snav[*]",
    ]

    def __init__(self, rcutfac, twojmax, element_profile, quadratic=False, **kwargs):
        """
        For more details on the parameters, please refer to the
        official documentation of LAMMPS.

        Notes:
            Despite this calculator uses compute sna(d)/atom command
            (http://lammps.sandia.gov/doc/compute_sna_atom.html), the
            parameter definition is in consistent with pair_style snap
            document (http://lammps.sandia.gov/doc/pair_snap.html).

        Args:
            rcutfac (float): The rcutfac in bispectrum coefficient calculations.
                The cutoff radius between element i and j are rcutfac * (R_i + R_j)
                where R_i and R_j are cutoff set for element i and j.
            twojmax (int): Band limit for bispectrum components.
            element_profile (dict): Parameters (cutoff radius 'r' and
                weight 'w') related to each element, e.g.,
                {'Na': {'r': 4.5, 'w': 0.9},
                 'Cl': {'r': 4.8, 'w': 3.0}}
            quadratic (bool): Whether including quadratic terms.
                Default to False.

        """
        self.rcutfac = rcutfac
        self.twojmax = twojmax
        self.element_profile = element_profile
        self.quadratic = quadratic
        super().__init__(**kwargs)

    @staticmethod
    def get_bs_subscripts(twojmax):
        """
        Method to list the subscripts 2j1, 2j2, 2j of bispectrum
        components.

        Args:
            twojmax (int): Band limit for bispectrum components.

        Returns:
            List of all subscripts [2j1, 2j2, 2j].

        """

        subs = itertools.product(range(twojmax + 1), repeat=3)

        filters = [lambda x: x[0] >= x[1], lambda x: x[2] >= x[0]]
        j_filter = [lambda x: x[2] in range(x[0] - x[1], min(twojmax, x[0] + x[1]) + 1, 2)]
        filters.extend(j_filter)
        for f in filters:
            subs = filter(f, subs)
        return list(subs)

    @property
    def n_bs(self):
        """
        Returns No. of bispectrum components to be calculated.

        """
        return len(self.get_bs_subscripts(self.twojmax))

    def _setup(self):
        def add_args(l):
            return l + compute_args if l.startswith("compute") else l

        compute_args = "1 0.99363 {} ".format(self.twojmax)
        el_in_seq = sorted(self.element_profile.keys(), key=lambda x: Element(x))
        cutoffs = [self.element_profile[e]["r"] * self.rcutfac for e in el_in_seq]
        weights = [self.element_profile[e]["w"] for e in el_in_seq]
        compute_args += " ".join([str(p) for p in cutoffs + weights])
        compute_args += " rmin0 0 quadraticflag {}".format(int(self.quadratic))
        CMDS = list(map(add_args, self._CMDS))
        CMDS[2] += " bzeroflag 0"
        CMDS[3] += " bzeroflag 0"
        CMDS[4] += " bzeroflag 0"
        dump_modify = "dump_modify 1 element "
        dump_modify += " ".join(str(e) for e in el_in_seq)
        CMDS.append(dump_modify)
        ALL_CMDS = self._COMMON_CMDS[:]
        ALL_CMDS[-1:-1] = CMDS
        input_file = "in.sna"
        with open(input_file, "w") as f:
            f.write(_pretty_input(ALL_CMDS))
        return input_file

    def _sanity_check(self, structure):
        elements = set(structure.symbol_set)
        sna_elements = self.element_profile.keys()
        return elements.issubset(sna_elements)

    def _parse(self):
        element = np.atleast_1d(_read_dump("dump.element", "unicode"))
        b = np.atleast_2d(_read_dump("dump.sna"))
        db = np.atleast_2d(_read_dump("dump.snad"))
        vb = np.atleast_2d(_read_dump("dump.snav"))
        return b, db, vb, element


class ElasticConstant(LMPStaticCalculator):
    """
    Elastic constant calculator.
    """

    _RESTART_CONFIG = {
        "internal": {"write_command": "write_restart", "read_command": "read_restart", "restart_file": "restart.equil"},
        "external": {"write_command": "write_data", "read_command": "read_data", "restart_file": "data.static"},
    }

    def __init__(
        self,
        ff_settings,
        potential_type="external",
        deformation_size=1e-6,
        jiggle=1e-5,
        lattice="bcc",
        alat=5.0,
        atom_type=1,
        maxiter=400,
        maxeval=1000,
        **kwargs,
    ):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for LAMMPS
                calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            potential_type (str): 'internal' indicates the internal potentials
                installed in lammps, 'external' indicates the external potentials
                outside of lammps.
            deformation_size (float): Finite deformation size. Usually range from
                1e-2 to 1e-8, to confirm the results not depend on it.
            jiggle (float): The amount of random jiggle for atoms to
                prevent atoms from staying on saddle points.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
            atom_type (int): The specified atom type if more than 1 species
                contained in the ff_settings.
            maxiter (float): The maximum number of iteration. Default to 400.
            maxeval (float): The maximum number of evaluation. Default to 1000.
        """
        self.ff_settings = ff_settings
        elements = ff_settings.elements
        self.write_command = self._RESTART_CONFIG[potential_type]["write_command"]
        self.read_command = self._RESTART_CONFIG[potential_type]["read_command"]
        self.restart_file = self._RESTART_CONFIG[potential_type]["restart_file"]
        self.deformation_size = deformation_size
        self.jiggle = jiggle
        self.lattice = lattice
        self.alat = alat
        self.num_species = len(elements)
        if isinstance(atom_type, str):
            atom_type = elements.index(atom_type) + 1
        self.atom_type = atom_type
        self.maxiter = maxiter
        self.maxeval = maxeval
        super().__init__(**kwargs)

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "elastic")

        with open(os.path.join(template_dir, "in.elastic"), "r") as f:
            input_template = f.read()
        with open(os.path.join(template_dir, "init.template"), "r") as f:
            init_template = f.read()
        with open(os.path.join(template_dir, "potential.template"), "r") as f:
            potential_template = f.read()
        with open(os.path.join(template_dir, "displace.template"), "r") as f:
            displace_template = f.read()

        input_file = "in.elastic"

        if isinstance(self.ff_settings, Potential):
            ff_settings = self.ff_settings.write_param()
        else:
            ff_settings = self.ff_settings

        with open(input_file, "w") as f:
            f.write(input_template.format(write_restart=self.write_command, restart_file=self.restart_file))
        with open("init.mod", "w") as f:
            f.write(
                init_template.format(
                    deformation_size=self.deformation_size,
                    jiggle=self.jiggle,
                    maxiter=self.maxiter,
                    maxeval=self.maxeval,
                    lattice=self.lattice,
                    alat=self.alat,
                    num_species=self.num_species,
                    atom_type=self.atom_type,
                    masses="\n".join(["mass {} {}".format(i + 1, i + 1) for i in range(self.num_species)]),
                )
            )
        with open("potential.mod", "w") as f:
            f.write(potential_template.format(ff_settings="\n".join(ff_settings)))
        with open("displace.mod", "w") as f:
            f.write(displace_template.format(read_restart=self.read_command, restart_file=self.restart_file))
        return input_file

    def calculate(self):
        """
        Calculate the elastic constant given Potential class.
        """
        with ScratchDir("."):
            input_file = self._setup()
            p = subprocess.Popen([self.LMP_EXE, "-in", input_file], stdout=subprocess.PIPE)
            stdout = p.communicate()[0]
            rc = p.returncode
            if rc != 0:
                error_msg = "LAMMPS exited with return code %d" % rc
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)
            result = self._parse()
        return result

    def _sanity_check(self, structure):
        """
        Check if the structure is valid for this calculation.

        """
        return True

    def _parse(self):
        """
        Parse results from dump files.

        """
        C11, C12, C44, bulkmodulus = np.loadtxt("elastic.txt")
        return C11, C12, C44, bulkmodulus


class LatticeConstant(LMPStaticCalculator):
    """
    Lattice Constant Relaxation Calculator.
    """

    def __init__(self, ff_settings, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
        """
        self.ff_settings = ff_settings
        super().__init__(**kwargs)

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "latt")

        with open(os.path.join(template_dir, "in.latt"), "r") as f:
            input_template = f.read()

        input_file = "in.latt"

        if isinstance(self.ff_settings, Potential):
            ff_settings = self.ff_settings.write_param()
        else:
            ff_settings = self.ff_settings

        with open(input_file, "w") as f:
            f.write(input_template.format(ff_settings="\n".join(ff_settings)))

        return input_file

    def _sanity_check(self, structure):
        """
        Check if the structure is valid for this calculation.

        """
        return True

    def _parse(self):
        """
        Parse results from dump files.

        """
        a, b, c = np.loadtxt("lattice.txt")
        return a, b, c


class NudgedElasticBand(LMPStaticCalculator):
    """
    NudgedElasticBand migration energy calculator.
    """

    def __init__(self, ff_settings, specie, lattice, alat, num_replicas=7, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
            num_replicas (int): Number of replicas to use.
        """
        self.ff_settings = ff_settings
        self.specie = specie
        self.lattice = lattice
        self.alat = alat
        self.num_replicas = num_replicas
        super().__init__(**kwargs)

    @staticmethod
    def get_unit_cell(specie, lattice, alat):
        """
        Get the unit cell from specie, lattice type and lattice constant.

        Args
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
        """
        if lattice == "fcc":
            unit_cell = Structure.from_spacegroup(
                sg="Fm-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "bcc":
            unit_cell = Structure.from_spacegroup(
                sg="Im-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "diamond":
            unit_cell = Structure.from_spacegroup(
                sg="Fd-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        else:
            raise ValueError("Lattice type is invalid.")

        return unit_cell

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "neb")

        with open(os.path.join(template_dir, "in.relax"), "r") as f:
            relax_template = f.read()
        with open(os.path.join(template_dir, "in.neb"), "r") as f:
            neb_template = f.read()

        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=self.alat)
        lattice_calculator = LatticeConstant(ff_settings=self.ff_settings)
        a, _, _ = lattice_calculator.calculate([unit_cell])[0]
        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=a)

        if self.lattice == "fcc":
            start_idx, final_idx = 95, 49
            scale_factor = [3, 3, 3]
        elif self.lattice == "bcc":
            start_idx, final_idx = 40, 14
            scale_factor = [3, 3, 3]
        elif self.lattice == "diamond":
            start_idx, final_idx = 7, 15
            scale_factor = [2, 2, 2]
        else:
            raise ValueError("Lattice type is invalid.")

        super_cell = unit_cell * scale_factor
        super_cell_ld = LammpsData.from_structure(super_cell, ff_elements=self.ff_settings.elements)
        super_cell_ld.write_file("data.supercell")

        with open("in.relax", "w") as f:
            f.write(
                relax_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    lattice=self.lattice,
                    alat=a,
                    specie=self.specie,
                    del_id=start_idx + 1,
                    relaxed_file="initial.relaxed",
                )
            )

        p = subprocess.Popen([self.LMP_EXE, "-in", "in.relax"], stdout=subprocess.PIPE)
        stdout = p.communicate()[0]

        rc = p.returncode
        if rc != 0:
            error_msg = "LAMMPS exited with return code %d" % rc
            msg = stdout.decode("utf-8").split("\n")[:-1]
            try:
                error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                error_msg += ", ".join(msg[error_line:])
            except Exception:
                error_msg += msg[-1]
            raise RuntimeError(error_msg)

        with open("in.relax", "w") as f:
            f.write(
                relax_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    lattice=self.lattice,
                    alat=a,
                    specie=self.specie,
                    del_id=final_idx + 1,
                    relaxed_file="final.relaxed",
                )
            )

        p = subprocess.Popen([self.LMP_EXE, "-in", "in.relax"], stdout=subprocess.PIPE)
        stdout = p.communicate()[0]

        rc = p.returncode
        if rc != 0:
            error_msg = "LAMMPS exited with return code %d" % rc
            msg = stdout.decode("utf-8").split("\n")[:-1]
            try:
                error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                error_msg += ", ".join(msg[error_line:])
            except Exception:
                error_msg += msg[-1]
            raise RuntimeError(error_msg)

        final_relaxed_struct = LammpsData.from_file("final.relaxed", atom_style="charge").structure

        lines = ["{}".format(final_relaxed_struct.num_sites)]

        for idx, site in enumerate(final_relaxed_struct):
            if idx == final_idx:
                idx = final_relaxed_struct.num_sites
            elif idx == start_idx:
                idx = final_idx
            else:
                idx = idx
            lines.append("{}  {:.3f}  {:.3f}  {:.3f}".format(idx + 1, site.x, site.y, site.z))

        with open("data.final_replica", "w") as f:
            f.write("\n".join(lines))

        input_file = "in.neb"

        with open(input_file, "w") as f:
            f.write(
                neb_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    start_replica="initial.relaxed",
                    final_replica="data.final_replica",
                )
            )

        return input_file

    def calculate(self):
        """
        Calculate the NEB barrier given Potential class.
        """
        with ScratchDir("."):
            input_file = self._setup()
            p = subprocess.Popen(
                [
                    "mpirun",
                    "--oversubscribe",
                    "-n",
                    str(self.num_replicas),
                    "lmp_mpi",
                    "-partition",
                    "{}x1".format(self.num_replicas),
                    "-in",
                    input_file,
                ],
                stdout=subprocess.PIPE,
            )
            stdout = p.communicate()[0]
            rc = p.returncode
            if rc != 0:
                error_msg = "LAMMPS exited with return code %d" % rc
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    logger.info(f"NudgedElasticBand error with message {msg}")
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)
            result = self._parse()
        return result

    def _sanity_check(self, structure):
        """
        Check if the structure is valid for this calculation.

        """
        return True

    def _parse(self):
        """
        Parse results from dump files.

        """
        with open("log.lammps") as f:
            lines = f.readlines()[-1:]
        migration_barrier = float(lines[0].split()[6])
        return migration_barrier


class DefectFormation(LMPStaticCalculator):
    """
    Defect formation energy calculator.
    """

    def __init__(self, ff_settings, specie, lattice, alat, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
        """
        self.ff_settings = ff_settings
        self.specie = specie
        self.lattice = lattice
        self.alat = alat
        super().__init__(**kwargs)

    @staticmethod
    def get_unit_cell(specie, lattice, alat):
        """
        Get the unit cell from specie, lattice type and lattice constant.

        Args
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
        """
        if lattice == "fcc":
            unit_cell = Structure.from_spacegroup(
                sg="Fm-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "bcc":
            unit_cell = Structure.from_spacegroup(
                sg="Im-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "diamond":
            unit_cell = Structure.from_spacegroup(
                sg="Fd-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        else:
            raise ValueError("Lattice type is invalid.")

        return unit_cell

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "defect")

        with open(os.path.join(template_dir, "in.defect"), "r") as f:
            defect_template = f.read()

        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=self.alat)
        lattice_calculator = LatticeConstant(ff_settings=self.ff_settings)
        a, _, _ = lattice_calculator.calculate([unit_cell])[0]
        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=a)

        if self.lattice == "fcc":
            idx, scale_factor = 95, [3, 3, 3]
        elif self.lattice == "bcc":
            idx, scale_factor = 40, [3, 3, 3]
        elif self.lattice == "diamond":
            idx, scale_factor = 7, [2, 2, 2]
        else:
            raise ValueError("Lattice type is invalid.")

        super_cell = unit_cell * scale_factor
        efs_calculator = EnergyForceStress(ff_settings=self.ff_settings)
        energy_per_atom = efs_calculator.calculate([super_cell])[0][0] / len(super_cell)

        super_cell_ld = LammpsData.from_structure(super_cell, ff_elements=self.ff_settings.elements)
        super_cell_ld.write_file("data.supercell")

        input_file = "in.defect"

        with open(input_file, "w") as f:
            f.write(
                defect_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    lattice=self.lattice,
                    alat=a,
                    specie=self.specie,
                    del_id=idx + 1,
                    relaxed_file="data.relaxed",
                )
            )

        return input_file, energy_per_atom, len(super_cell) - 1

    def calculate(self):
        """
        Calculate the vacancy formation given Potential class.
        """
        with ScratchDir("."):
            input_file, energy_per_atom, num_atoms = self._setup()
            p = subprocess.Popen([self.LMP_EXE, "-in", input_file], stdout=subprocess.PIPE)
            stdout = p.communicate()[0]

            rc = p.returncode
            if rc != 0:
                error_msg = "LAMMPS exited with return code %d" % rc
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)
            defect_energy, _, _ = self._parse()
        defect_formation_energy = defect_energy - energy_per_atom * num_atoms

        return defect_formation_energy

    def _sanity_check(self, structure):
        return True

    def _parse(self):
        energy = float(np.loadtxt("energy.txt"))
        force = _read_dump("force.dump")
        stress = np.loadtxt("stress.txt")
        return energy, force, stress


class LammpsPotential(Potential, abc.ABC):  # type: ignore
    """
    Lammps compatible potentials that call lammps executable for
    energy/force/stress calculations
    """

    def predict_efs(self, structure):
        """
        Predict energy, forces and stresses of the structure.

        Args:
            structure (Structure): Pymatgen Structure object.

        Returns:
            energy, forces, stress
        """
        calculator = EnergyForceStress(self)
        energy, forces, stress = calculator.calculate(structures=[structure])[0]
        return energy, forces, stress



"""
Get data from url
"""
from io import StringIO

import requests

import pandas as pd

from maml import BaseDataSource


class URLSource(BaseDataSource):
    """
    Load raw data from a URL, e.g., figshare.
    """

    def __init__(self, fmt: str = "csv", read_kwargs=None):
        """

        Args:
            fmt: The format to read the raw data. It should be something supported by pandas.read_<something>.
            Defaults to "csv".
            read_kwargs: kwargs passed to the read_<something> command.
        """
        self.fmt = fmt
        self.read_kwargs = read_kwargs or {}

    def get(self, url: str) -> pd.DataFrame:  # type: ignore
        """
        Get url data source

        Args:
            url: URL to obtain raw data from.

        Returns:
            pd.DataFrame
        """
        raw = requests.get(url).text
        read_func = getattr(pd, "read_%s" % self.fmt)
        return read_func(StringIO(raw), **self.read_kwargs)


class FigshareSource(URLSource):
    """
    Load data from figshare.
    """

    def get(self, file_id: str) -> pd.DataFrame:  # type: ignore
        """
        Get data from Figshare
        Args:
            file_id: file id

        Returns:
            data frame
        """

        url = "https://ndownloader.figshare.com/files/%s" % file_id
        return super().get(url)



"""Signal processing utils"""
from math import ceil, floor
from typing import Union, Tuple, Callable

import numpy as np

from scipy import fft
from scipy import signal

from monty.dev import requires

try:
    import tftb
except ImportError:
    tftb = None


def fft_magnitude(z: np.ndarray) -> np.ndarray:
    """
    Dicrete Fourier Transform the signal z and return
    the magnitude of the  coefficents
    Args:
        z (np.ndarray): 1D signal array
    Returns: 1D magnitude
    """
    return np.absolute(fft.fft(z))


def spectrogram(z: np.ndarray, return_time_freq: bool = False) -> Union[Tuple, np.ndarray]:
    """
    The spectrogram of the signal
    Args:
        z (np.ndarray): 1D signal array
        return_time_freq (bool): whether to return time and frequency
    Returns: 2D spectrogram
    """
    nx = len(z)
    nsc = floor(nx / 4.5)  # use matlab values
    nov = floor(nsc / 2)
    nfft = max(256, 2 ** ceil(np.log2(nsc)))
    freq, time, s = signal.spectrogram(z, window=signal.get_window("hann", nsc), noverlap=nov, nfft=nfft)
    if return_time_freq:
        return freq, time, s
    return s


def cwt(z: np.ndarray, widths: np.ndarray, wavelet: Union[str, Callable] = "morlet2", **kwargs) -> np.ndarray:
    """
    The scalogram of the signal
    Args:
        z (np.ndarray): 1D signal array
        widths (np.ndarray): wavelet widths
        wavelet (str): wavelet name
    Returns: 2D scalogram
    """
    if isinstance(wavelet, str):
        wavelet_func = getattr(signal, wavelet)
    else:
        wavelet_func = wavelet
    return np.absolute(signal.cwt(z, wavelet_func, widths=widths, **kwargs))


@requires(tftb is not None, "Requires installation of tftb package")
def wvd(z: np.ndarray, return_all: bool = False) -> Union[Tuple, np.ndarray]:
    """
    Wigner Ville Distribution calculator
    Args:
        z (np.ndarray): signal 1D
        return_all (bool): whether to return time and freq info, default
            only return the wvd information
    Returns: NxN wvd matrix
    """
    tfr = tftb.processing.WignerVilleDistribution(z)
    (
        res,
        f1,
        f2,
    ) = tfr.run()
    if return_all:
        return res, f1, f2
    return res


AVAILABLE_SP_METHODS = {"fft_magnitude": fft_magnitude, "spectrogram": spectrogram, "cwt": cwt, "wvd": wvd}


def get_sp_method(sp_method: Union[str, Callable]) -> Callable:  # type: ignore
    """
    Providing a signal processing method name return the callable
    Args:
        sp_method (str): name of the sp function
    Returns: callable for signal processing
    """
    if isinstance(sp_method, str):
        try:
            return AVAILABLE_SP_METHODS[sp_method]
        except KeyError:
            raise KeyError(f"{sp_method} is not in available methods: " f"{AVAILABLE_SP_METHODS.keys()}")
    else:
        return sp_method



# coding: utf-8
# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides SNAP interatomic potential class."""

import re

import numpy as np
from sklearn.linear_model import LinearRegression

from monty.io import zopen

from maml import SKLModel
from maml.describers import BispectrumCoefficients
from maml.utils import pool_from, convert_docs, check_structures_forces_stresses
from ._lammps import LammpsPotential


class SNAPotential(LammpsPotential):
    """
    This class implements Spectral Neighbor Analysis Potential.
    """

    pair_style = "pair_style        snap"
    pair_coeff = "pair_coeff        * * {coeff_file} {param_file} {elements}"

    def __init__(self, model, name=None):
        """
        Initialize the SNAPotential Potential with atomic describers
        and models, which are used to generate the Bispectrum coefficients
        features for structures and to train the parameters.

        Args:
            model (LinearModel): Model to perform supervised learning with
                atomic descriptos as features and properties as targets.
            name (str): Name of force field.
        """
        self.name = name if name else "SNAPotential"
        self.model = model
        self.elements = self.model.describer.elements

    def train(
        self, train_structures, train_energies, train_forces, train_stresses=None, include_stress=False, **kwargs
    ):
        """
        Training data with models.

        Args:
            train_structures ([Structure]): The list of Pymatgen Structure object.
                energies ([float]): The list of total energies of each structure
                in structures list.
            train_energies ([float]): List of total energies of each structure in
                structures list.
            train_forces ([np.array]): List of (m, 3) forces array of each
                structure with m atoms in structures list. m can be varied with
                each single structure case.
            train_stresses (list): List of (6, ) virial stresses of each
                structure in structures list.
            include_stress (bool): Whether to include stress components.
        """
        train_structures, train_forces, train_stresses = check_structures_forces_stresses(
            train_structures, train_forces, train_stresses
        )
        train_pool = pool_from(train_structures, train_energies, train_forces, train_stresses)
        _, df = convert_docs(train_pool, include_stress=include_stress)
        ytrain = df["y_orig"] / df["n"]
        xtrain = self.model.describer.transform(train_structures)
        self.model.fit(features=xtrain, targets=ytrain, **kwargs)

    def evaluate(self, test_structures, test_energies, test_forces, test_stresses=None, include_stress=False):
        """
        Evaluate energies, forces and stresses of structures with trained
        machinea learning potentials.

        Args:
            test_structures ([Structure]): List of Pymatgen Structure Objects.
            test_energies ([float]): List of DFT-calculated total energies of
                each structure in structures list.
            test_forces ([np.array]): List of DFT-calculated (m, 3) forces of
                each structure with m atoms in structures list. m can be varied
                with each single structure case.
            test_stresses (list): List of DFT-calculated (6, ) viriral stresses
                of each structure in structures list.
            include_stress (bool): Whether to include stress components.
        """
        test_structures, test_forces, test_stresses = check_structures_forces_stresses(
            test_structures, test_forces, test_stresses
        )
        predict_pool = pool_from(test_structures, test_energies, test_forces, test_stresses)
        _, df_orig = convert_docs(predict_pool, include_stress=include_stress)
        _, df_predict = convert_docs(pool_from(test_structures), include_stress=include_stress)
        outputs = self.model.predict_objs(objs=test_structures)
        df_predict["y_orig"] = df_predict["n"] * outputs

        return df_orig, df_predict

    def write_param(self):
        """
        Write parameter and coefficient file to perform lammps calculation.
        """
        param_file = "{}.snapparam".format(self.name)
        coeff_file = "{}.snapcoeff".format(self.name)
        model = self.model
        describer = self.model.describer
        profile = describer.element_profile
        ne = len(self.elements)
        nbc = len(describer.subscripts)
        if describer.quadratic:
            nbc += int((1 + nbc) * nbc / 2)

        coeff_lines = []
        coeff_lines.append("{} {}".format(ne, nbc + 1))
        for element, coeff in zip(self.elements, np.split(model.model.coef_, ne)):
            coeff_lines.append("{} {} {}".format(element, profile[element]["r"], profile[element]["w"]))
            coeff_lines.extend([str(c) for c in coeff])
        with open(coeff_file, "w") as f:
            f.write("\n".join(coeff_lines))

        param_lines = []
        keys = ["rcutfac", "twojmax"]
        param_lines.extend(["{} {}".format(k, getattr(describer, k)) for k in keys])
        param_lines.extend(["rfac0 0.99363", "rmin0 0"])
        param_lines.append("quadraticflag {}".format(int(describer.quadratic)))
        param_lines.append("bzeroflag 0")
        with open(param_file, "w") as f:
            f.write("\n".join(param_lines))

        pair_style = self.pair_style
        pair_coeff = self.pair_coeff.format(
            elements=" ".join(self.elements), coeff_file=coeff_file, param_file=param_file
        )
        ff_settings = [pair_style, pair_coeff]
        return ff_settings

    @staticmethod
    def from_config(param_file, coeff_file, **kwargs):
        """
        Initialize potentials with parameters file and coefficient file.

        Args:
            param_file (str): The file storing the configuration of potentials.
            coeff_file (str): The file storing the coefficients of potentials.

        Return:
            SNAPotential.
        """
        with open(coeff_file) as f:
            coeff_lines = f.readlines()
        coeff_lines = [line for line in coeff_lines if not line.startswith("#")]
        element_profile = {}
        ne, nbc = coeff_lines[0].split()
        ne, nbc = int(ne), int(nbc)
        for n in range(ne):
            specie, r, w = coeff_lines[1 + n * (nbc + 1)].split()
            r, w = float(r), float(w)
            element_profile[specie] = {"r": r, "w": w}

        rcut_pattern = re.compile(r"rcutfac (.*?)\n", re.S)
        twojmax_pattern = re.compile(r"twojmax (\d*)\n", re.S)
        quadratic_pattern = re.compile(r"quadraticflag (.*?)(?=\n|$)", re.S)

        with zopen(param_file, "rt") as f:
            param_lines = f.read()

        rcut = float(rcut_pattern.findall(param_lines)[-1])
        twojmax = int(twojmax_pattern.findall(param_lines)[-1])
        if quadratic_pattern.findall(param_lines):
            quadratic = bool(int(quadratic_pattern.findall(param_lines)[-1]))
        else:
            quadratic = False

        describer = BispectrumCoefficients(
            rcutfac=rcut, twojmax=twojmax, element_profile=element_profile, quadratic=quadratic, pot_fit=True
        )
        model = SKLModel(model=LinearRegression(), describer=describer, **kwargs)
        coef = np.array(
            np.concatenate([coeff_lines[(2 + nbc * n + n) : (2 + nbc * (n + 1) + n)] for n in range(ne)]),
            dtype=np.float,
        )
        model.model.coef_ = coef
        model.model.intercept_ = 0
        snap = SNAPotential(model=model)
        return snap



# coding: utf-8
# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This package contains PotentialMixin classes representing Interatomic Potentials."""

from ._base import PotentialMixin, Potential  # noqa
from ._gap import GAPotential  # noqa
from ._lammps import (  # noqa
    LMPStaticCalculator,
    EnergyForceStress,  # noqa
    SpectralNeighborAnalysis,
    ElasticConstant,  # noqa
    LatticeConstant,
    NudgedElasticBand,
    DefectFormation,  # noqa
    get_default_lmp_exe,  # noqa
)
from ._mtp import MTPotential  # noqa
from ._nnp import NNPotential  # noqa
from ._snap import SNAPotential  # noqa

__all__ = [
    "Potential",
    "GAPotential",
    "MTPotential",
    "SNAPotential",
    "NNPotential",
    "LMPStaticCalculator",
    "EnergyForceStress",
    "SpectralNeighborAnalysis",
    "ElasticConstant",
    "LatticeConstant",
    "NudgedElasticBand",
    "DefectFormation",
    "get_default_lmp_exe",
]



# coding: utf-8
# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

from setuptools import setup, find_packages

setup(
    name="maml",
    packages=find_packages(),
    version='0.0.4',
    install_requires=["numpy", "scipy", "monty",
                      "scikit-learn", "pandas", "pymatgen", "tqdm"],
    extras_requires={"maml.apps.symbolic._selectors_cvxpy": ["cvxpy"],
        "tensorflow": ["tensorflow>=2"],
        "tensorflow with gpu": ["tensorflow-gpu>=2"]},
    author="Materials Virtual Lab",
    author_email="ongsp@eng.ucsd.edu",
    maintainer="Shyue Ping Ong",
    maintainer_email="ongsp@eng.ucsd.edu",
    url="http://www.materialsvirtuallab.org",
    license="BSD",
    description="maml is a machine learning library for materials science.",
    long_description="""Maml, acronym for MAterials Machine Learning and pronounced mammal, is a machine learning 
    library for materials science. It builds on top of pymatgen (Python Materials Genomics) materials analysis library
and well-known machine learning/deep learning libraries like scikit-learn, Keras and Tensorflow. The aim is to link the power of both kinds of 
libraries for rapid experimentation and learning of materials data.""",
    long_description_content_type='text/markdown',
    keywords=["materials", "science", "deep", "learning"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    include_package_data=True,
    package_data={'maml': ['describers/data/*.json', 'describers/data/megnet_models/*.json'
        'describers/data/megnet_mdoels/*.hdf5']}
)



"""
ValueProfile return values according to certain settings. For
example, one can design a linearly increasing value profile,
a sinusoidal profile or a constant profile, depending on the
step, and previous values.
"""
from typing import Optional

import numpy as np


class ValueProfile:
    """
    Base class for ValueProfile. The base class has the following methods
    methods:
        increment_step(self): add one to step
        get_value(self): abstract method that return the value
    """

    def __init__(self, max_steps: Optional[int] = None, **kwargs):
        """
        Initializer for ValueProfile. It sets step to 0
        Args:
            **kwargs: any kwargs
        """
        self.step = 0
        if max_steps is None:
            max_steps = np.infty
        self.max_steps = max_steps

    def increment_step(self):
        """
        increase step atttribute by one
        """
        self.step += 1
        if self.step > self.max_steps:
            raise RuntimeError("Step exceeding maximum")

    def get_value(self) -> float:
        """
        abstract method that returns the current value
        Returns: value float

        """
        raise NotImplementedError()


class ConstantValue(ValueProfile):
    """
    Return constant value
    """

    def __init__(self, value: float, **kwargs):
        """
        Initialize constant profile
        Args:
            value (float): constant value
            **kwargs:
        """
        self.value = value
        super().__init__(**kwargs)

    def get_value(self) -> float:
        """
        Return constant value
        """

        return self.value


class LinearProfile(ValueProfile):
    """
    LinearProfile by setting starting value and the rate of
    value change. The profile can be initialized either by
    [value_start, value_end, max_step] or [value_start, rate]
    """

    def __init__(self, value_start: float, value_end: float = 0.0, max_steps: int = 100, **kwargs):

        """

        Args:
            value_start (float):  start value
            value_end (float): end value, optional
            max_step (int): number of steps, optional
            rate (float): rate of value change
            **kwargs: captures anything else
        """

        self.value_start = value_start
        rate = kwargs.get("rate", None)
        if rate is None:
            self.rate = (value_end - self.value_start) / max_steps
        else:
            self.rate = rate
            max_steps = np.infty
        super().__init__(max_steps=max_steps, **kwargs)

    def get_value(self) -> float:
        """
        Get LinearProfile value
        Returns: float
        """
        return self.value_start + self.step * self.rate



"""MAML models"""
from maml.base import SKLModel, KerasModel  # noqa
from .dl import WeightedSet2Set, WeightedAverageLayer, AtomSets, MLP  # noqa


__all__ = ["SKLModel", "KerasModel", "AtomSets", "MLP", "WeightedSet2Set", "WeightedAverageLayer"]



"""
The core classes for XAS prediction app
"""
import os
import warnings

import numpy as np
import pandas as pd
from pymatgen.core.spectrum import Spectrum


class XANES(Spectrum):
    """
    Basic XANES object
    """

    XLABEL = "Energy"
    YLABEL = "Intensity"

    def __init__(self, x, y, absorption_specie, edge, structure=None, e0=None, **kwargs):
        """
        Args:
            x (list/np.array): A sequence of x-ray energies in eV
            y (list/np.array): A sequence of mu(E)
            absorption_specie (str): Specie associated with the XANES.
            edge (str): Spectrum edge, e.g. K, L1, L2, etc
            structure (Structure): Structure associated with the XANES. Pymatgen structure object
            e0 (float): Edge energy associated with the spectrum
            **kwargs: Other attributes of the spectrum object.
        """
        super().__init__(x, y, absorption_specie, edge, structure, e0, **kwargs)
        self.absorption_specie = absorption_specie
        self.edge = edge
        self.structure = structure
        if e0:
            self.e0 = e0
        else:
            warning_msg = "Edge energy is determined with maximum derivative. Using this e0 with caution."
            warnings.warn(warning_msg)
            self.e0 = self.x[np.argmax(np.gradient(self.y) / np.gradient(self.x))]

        for (field, value) in kwargs.items():
            setattr(self, field, value)

    def __str__(self):
        if self.structure:
            return "%s %s Edge for %s: %s" % (
                self.absorption_specie,
                self.edge,
                self.structure.composition.reduced_formula,
                super().__str__(),
            )
        return "%s %s Edge for %s" % (self.absorption_specie, self.edge, super().__str__())

    @staticmethod
    def from_K_XANES_MP_tsv(tsv_file, header=3, absorption_specie=None, **kwargs):
        """
        Args:
            tsv_file: tsv file contains spectrum infor. downloaded from the Materials Project website
            header: number of header row to skip, default to 3
            absorption_specie (str): Specie associated with the XANES. If not given, will be derived using the
                                    edge energy.
            **kwargs: Other attributes of the spectrum object.

        Returns:
            XANES object generated from tsv file

        """

        spectrum_sample = pd.read_csv(tsv_file, sep="\t", header=header)
        spectrum_energy = spectrum_sample["E (eV)"]
        spectrum_mu = spectrum_sample["mu"]

        if absorption_specie is None:
            warning_msg = "Absorption specie is derived using the edge energy determined with maximum derivative."
            warnings.warn(warning_msg)
            edge_energy = spectrum_energy[np.argmax(np.gradient(spectrum_mu) / np.gradient(spectrum_energy))]
            edge_vs_atomNo_filepath = os.path.join(os.path.dirname(__file__), "data", "Edge_vs_atomNo.csv")
            edge_vs_atomNo_df = pd.read_csv(edge_vs_atomNo_filepath)
            kedge_vs_atomNo_df = edge_vs_atomNo_df[["Element", "K Edge"]]
            kedge_vs_atomNo_df["K_edge_difference"] = kedge_vs_atomNo_df["K Edge"].apply(
                lambda x: np.abs(x - edge_energy)
            )
            derived_abs_specie = kedge_vs_atomNo_df.sort_values("K_edge_difference").iloc[0]["Element"]
            return XANES(spectrum_energy, spectrum_mu, absorption_specie=derived_abs_specie, edge="K", **kwargs)

        return XANES(spectrum_energy, spectrum_mu, absorption_specie=absorption_specie, edge="K", **kwargs)



"""
MAML describers base classes
"""
import abc
import logging
import tempfile
from typing import Any, List

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_memory

from joblib import cpu_count, Parallel, delayed
from tqdm import tqdm

from monty.json import MSONable

from maml.utils import feature_dim_from_test_system
from ._feature_batch import get_feature_batch

_ALLOWED_DATA = ("number", "structure", "molecule", "spectrum")

_DESCRIBER_TYPES = ["composition", "site", "structure", "general", "band_structure", "spectrum"]

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BaseDescriber(BaseEstimator, TransformerMixin, MSONable, metaclass=abc.ABCMeta):
    """
    Base class for a Describer. A describers converts an object to a descriptor,
    typically a numerical representation useful for machine learning.
    The output for the describers can be a single DataFrame/numpy.ndarray or
    a list of DataFrame/numpy.ndarray.
    """

    def __init__(self, **kwargs):
        """
        Base estimator with the following allowed keyword args

            memory (bool/str/joblib.Memory): The path or Memory for caching the computational
                results, default None means no cache.
            verbose (bool): Whether to show the progress of feature calculations.
            n_jobs (int): The number of parallel jobs. 0 means no parallel computations.
                If this value is set to negative or greater than the total cpu
                then n_jobs is set to the number of cpu on system.
            feature_batch (str): method to batch a list of features into one

        Args:
            **kwargs: keyword args that contain possibly memory (str/joblib.Memory),
                verbose (bool), n_jobs (int)
        """
        allowed_kwargs = ["memory", "verbose", "n_jobs", "feature_batch"]

        for k, v in kwargs.items():
            if k not in allowed_kwargs:
                raise TypeError("%s not allowed as kwargs" % (str(k)))
        memory = kwargs.get("memory", None)
        if isinstance(memory, bool):
            memory = tempfile.mkdtemp()
            logger.info("Created temporary directory %s" % memory)
        verbose = kwargs.get("verbose", False)
        n_jobs = kwargs.get("n_jobs", 0)

        self.memory = check_memory(memory)
        self.verbose = verbose
        # find out the number of parallel jobs
        if (n_jobs < 0) or (n_jobs > cpu_count()):
            n_jobs = cpu_count()
            logger.info(f"Using {n_jobs} jobs for computation")
        self.n_jobs = n_jobs
        self.feature_batch = get_feature_batch(kwargs.get("feature_batch", None))

    def fit(self, x: Any, y: Any = None) -> "BaseDescriber":
        """
        Place holder for fit API

        Args:
            x: Any inputs
            y: Any outputs

        Returns: self

        """
        return self

    def transform_one(self, obj: Any) -> np.ndarray:
        """
        Transform an object.
        """
        raise NotImplementedError

    def transform(self, objs: List[Any]) -> Any:
        """
        Transform a list of objs. If the return data is DataFrame,
        use df.xs(index, level='input_index') to get the result for the i-th object.

        Args:
            objs (list): A list of objects.
        Returns:
            One or a list of pandas data frame/numpy ndarray
        """
        cached_transform_one = self.memory.cache(_transform_one)

        if self.verbose:
            objs = tqdm(objs)

        # run the featurizer
        if self.n_jobs == 0:
            features = [cached_transform_one(self, obj) for obj in objs]
        else:
            features = Parallel(n_jobs=self.n_jobs)(delayed(cached_transform_one)(self, obj) for obj in objs)

        multi_output = self._is_multi_output()
        if not multi_output:
            features = [features]
        batched_features = [self.feature_batch(i) for i in list(*zip(features))]  # type: ignore
        return batched_features if multi_output else batched_features[0]

    def _is_multi_output(self) -> bool:
        tags = self._get_tags()
        multi_output = tags["multioutput"]  # this is from BaseEstimator
        return multi_output

    def clear_cache(self):
        """
        Clear cache
        """
        if self.memory.location is not None:
            self.memory.clear()

    @property
    def feature_dim(self):
        """
        Feature dimension, useful when certain models need to specif
        the feature dimension, e.g., MLP models.
        """
        return feature_dim_from_test_system(self)


def _transform_one(describer: BaseDescriber, obj: Any) -> np.ndarray:
    """
    A wrapper to make a pure function.

    Args:
        describer (BaseDescriber): a describers

    Returns:
        np.ndarray
    """
    return describer.transform_one(obj)


class DummyDescriber(BaseDescriber):
    """
    Dummy Describer that does nothing
    """

    def transform_one(self, obj: Any):
        """
        Does nothing but return the original features

        Args:
            obj: Any inputs

        Returns: Any outputs

        """
        return obj


class SequentialDescriber(Pipeline):
    """
    A thin wrapper of sklearn Pipeline
    """

    def __init__(self, describers: List, **kwargs):
        """
        Put a list of describers into one pipeline
        Args:
            describers (list): a list of describers that will be applied
                consecutively
            **kwargs:
        """
        steps = [(i.__class__.__name__, i) for i in describers]
        super().__init__(steps, **kwargs)


def describer_type(dtype: str):
    """
    Decorate to set describers class type

    Args:
        dtype (str): describers type, e.g., site, composition, structure etc.

    Return:
        wrapped class
    """

    def wrapped_describer(klass):
        klass.describer_type = dtype
        return klass

    return wrapped_describer
